#ifndef _ARITH_H
#define _ARITH_H      1
                           //flags that this file has been included
#include <stream.h>
#include <builtin.h>

#define USE_DMOD

#ifdef USE_DMOD
#define BIGPRIME 134217689   // = 2^27-39
#define xmod(a,p) dmod((a),(p))
#define xmodmul(a,b,p) dmodmul((a),(b),(p))
#else
#define BIGPRIME 92681 //This is the largest p such that (p/2)^2 < 2^31.
#define xmod(a,p) mod((a),(p))
#define xmodmul(a,b,p) mod((a)*(b),(p))
#endif

// The following achieve modular reduction and multiplication using doubles, 
// which allows a prime modulus of just under 2^27 to be used; without, the
// largest is under 2^16.5

inline long dmod(long a, long m)
{
  double xa = a;
  return (long)(xa-(m*floor((xa/m)+0.5)));
}

inline long dmodmul(long a, long b, long m)
{ double c = (double)a * (double)b;
  c-=m*floor((c/m)+0.5);
//  cout << "dmodmul returns " << (long)c << "\n";
  return (long)c;
}


/* Prime number procs; adapted from Pari  */

typedef unsigned char *byteptr;
extern byteptr diffptr;
extern long NPRIMES, BIGGESTPRIME;

class longlist {
friend class longvar;
public:
        long length;        
private:
        long *items;  
public:
        longlist(long n=0);
        ~longlist();
        longlist(const longlist& l);
        longlist(const longlist& l, long n);
        longlist& operator=(const longlist&l);
        long& get(long i);
        long locate(long num) const;
        long& operator[](long i);
        long operator()(long i) const;
        void set(long i, long val);
        void truncate(long newlength);
        void output(ostream& os) const;
        friend inline ostream& operator<<(ostream& os, const longlist& l)
              {l.output(os);return os;}
};

class longvar {
public:
        long index;        /* current index */
private:
        long* values;
        long maxindex;     /* max index */
public:
        longvar(const longlist& l) 
          {maxindex=l.length-1; index=0; values=l.items;}
        void init(const longlist& l) 
          {maxindex=l.length-1; index=0; values=l.items;}
        long next();
        long operator++(int);   // postfix version
        long operator++();      // prefix version
        int ok() const {return index<=maxindex;}
        int more() const {return index<maxindex;}
        long value() const {return *values;}
        operator long() const {return *values;}
};

/* Usage of longvar: to loop through a longlist "alist" (which might be empty):
   either of

   for(longvar avar(alist); avar.ok(); avar++) {long a=avar; ... ;}
   for(avar.init(alist); avar.ok(); avar++) {long a=avar; ... ;}  
   // second form IFF avar already exists

   Either way, the argument, alist, must be an instance of an existing
   longlist, and must not be a function returning a longlist (else
   unflagged runtime error will follow as the temporary is deleted)
*/

long prime_number(long); //forward declaration, see later 

class primevar {
public:
        long value;        /* current value */
        long index;        /* current index */
private:
        byteptr ndiff;     /* pointer to next diff*/
        long maxindex;     /* max index */
public:
        primevar(long max=NPRIMES, long i=1) 
          {
	    maxindex=max; if(maxindex>NPRIMES) maxindex=NPRIMES; 
	    index=i;value=prime_number(i);ndiff=diffptr+i;
	  }
        void init(long max=NPRIMES, long i=1) 
          {
	    maxindex=max; if(maxindex>NPRIMES) maxindex=NPRIMES; 
	    index=i;value=prime_number(i);ndiff=diffptr+i;
	  }
        long next();
        long operator++(int);  // postfix version
        long operator++();     // prefix  version
        int ok() const {return index<=maxindex;}
        int more() const {return index<maxindex;}
        operator long() const {return value;}
};

/* Usage of primevar: 
---To loop through first n primes:

   long p; 
   for(primevar pr(n); pr.ok(); pr++) {p=pr; ... ;}  // or:
   for(pr.init(n); pr.ok(); pr++) {p=pr; ... ;}  // iff pr is existing primevar

---To loop through all primes:
   
   primevar pr; //or for existing primevar, pr.init();
   long p;
   while(pr.ok()) {p=pr; pr=++; ...;}

*/

long initprimes(long maxnum);  /* initializes global variable diffptr
                                 MUST be called before any other function
                                 which uses primes */
//long prime_number (long n);/* returns n'th prime from global list */
longlist primes (long n);  /* returns list of first n primes */
long primdiv(long);        /* returns first prime divisor */
longlist pdivs(long);      /* list of prime divisors */

longlist posdivs(long, const longlist& plist);  // all positive divisors
inline longlist posdivs(long n)
{
  const longlist& plist = pdivs(n);
  return posdivs(n,plist);
}

longlist alldivs(long, const longlist& plist);  // absolutely all divisors
inline longlist alldivs(long n)
{
  const longlist& plist = pdivs(n);
  return alldivs(n,plist);
}

longlist sqdivs(long, const longlist& plist);   //  divisors whose square divides
inline longlist sqdivs(long n)
{
  const longlist& plist = pdivs(n);
  return sqdivs(n,plist);
}

longlist sqfreedivs(long, const longlist& plist); // square-free divisors
inline longlist sqfreedivs(long n)
{
  const longlist& plist = pdivs(n);
  return sqfreedivs(n,plist);
}


long mod(long a, long modb); /* modulus in range plus or minus half mod */
long posmod(long a, long modb); /* ordinary modulus, but definitely positive */

long bezout(long, long, long&, long&);
int intbezout(int aa, int bb, int& xx, int& yy);

long invmod(long, long);
int modrat(long, long, float, long&, long&);

long val(long factor, long number); // order of factor in number
int div(long factor,long number);  // test for divisibility
int ndiv(long factor,long number); // test for non-divisibility

long m1pow(long a);
long chi2(long a);
long chi4(long a);
long hilbert(long a, long b);
long legendre(long a, long b);
long kronecker(long d, long n);
long log2(long n,long& e);       // returns 1,e if n=2^e, else 0 

// Implementations of inline functions

inline long& longlist::get(long i) 
{
  if ((i>=0)&&(i<length)) return items[i]; 
  else{cerr<<"bad index "<<i<<" in longlist!\n"; return items[0];}
}

inline long& longlist::operator[](long i) {return this->get(i);}

inline long longlist::operator()(long i) const 
{
  if ((i>=0)&&(i<length)) return items[i]; 
  else{cerr<<"bad index "<<i<<" in longlist!\n"; return 0;}
}

inline void longlist::set(long i, long val) 
{
  if ((i>=0)&&(i<length)) items[i]=val; 
  else {cerr<<"bad index "<<i<<" in longlist!\n";}
}

inline void longlist::truncate(long newlength) 
{ 
  if (newlength<length) length = newlength;
}
  
int startswith(const longlist& a, const longlist& b, long l, long from);

inline long longvar::next() 
{
  if ((index++)<maxindex) {values++; return 1;}
  else return 0;
}

inline long longvar::operator++()       // prefix
{
  if ((index++)<=maxindex) {values++; return *values;}
  else return 0;
}

inline long longvar::operator++(int x)  // postfix
{
  if ((index++)<=maxindex) return *values++;
  else return 0;
}

inline long primevar::next() 
{
  if ((index++)<maxindex) {value+=*ndiff++; return 1;}
  else return 0;
}

inline long primevar::operator++()   // prefix
{
  if ((index++)<=maxindex) {long ans=value; value+=*ndiff++;
                            return ans;}
  else return 0; 
}

inline long primevar::operator++(int x)  // postfix (same)
{
  if ((index++)<=maxindex) {long ans=value; value+=*ndiff++;
                            return ans;}
  else return 0; 
}

inline int div(long factor,long number) 
{ 
  return (factor==0) ? (number==0) : (number%factor==0);
}

inline int ndiv(long factor,long number) 
{
  return !div(factor,number);
}

inline long m1pow(long a)
{ 
  return (a%2 ?  -1 : +1);
}
 

#endif
