#include "marith.h"
#include "conic.h"

//#define DEBUG_CONIC
//#define DEBUG_CONIC_2
//#define TRACE_FACTORIZATION
//#define DEBUG_REDUCE

//#define METHOD 0 // simple recursion, no reduction
//#define METHOD 1 // recursion with algebraic reduction (Denis Simon)
#define METHOD 2 // recursion with lattice reduction
//#define METHOD 3 // obsolete

void show_xyz(const bigint& x, const bigint& y, const bigint& z)
{
  cout << "(x:y:z) = ("<<x<<":"<<y<<":"<<z<<")\n";
}

bigint cancel1(bigint& x, bigint& y, bigint& z)
     // cancels common factors only
{
  bigint g=gcd(x,y);
  if(!is_one(g)) 
    {
      g=gcd(g,z);
      if(!is_one(g)) {x/=g; y/=g; z/=g;}
    }
  return g;
}

void cancel(bigint& x, bigint& y, bigint& z)
     // cancels common factors and leaves z>=0 or z=0 and x>=0
{
  cancel1(x,y,z);
  if(is_positive(z)) return;
  if(is_negative(z)) {x.negate(); y.negate(); z.negate(); return;}
  if(is_positive(y)) return;
  if(is_negative(y)) {x.negate(); y.negate(); return;}
  if(is_negative(x)) {x.negate();}
  return;
}

int solve_conic(const bigint& a, const bigint& b, const bigint& c, const bigint& d,
                bigint& x, bigint& y, bigint& z)
     // Solves axx+bxz+czz=dyy for (x,y,z) not (0,0,0) and returns 1
     // or returns 0 if not possible
     // Should have a, c, d, bb-4ac non-zero
{
  // All this function does is complete the square (if nec.) and pass to the next
  bigint aa, bb, xx, yy, zz, fac;
  if((is_one(a))&&(is_zero(b))) {aa=-c; bb=d; fac=1;}
  else {aa=sqr(b)-4*a*c; bb=a*d; fac=2;}
  bigint a1,a2,b1,b2;
  bigintArray aplist, bplist;
  sqfdecomp(aa,a1,a2,aplist);  // aa=a1*a2^2
  sqfdecomp(bb,b1,b2,bplist);  // bb=b1*b2^2
  if(!testlocsol(a1,aplist,b1,bplist)) return 0;
  if(solve_conic_diag(a1,aplist,b1,bplist,xx,yy,zz))
    {
      xx*=(a2*b2); yy*=a2; zz*=b2; 
      x = xx-b*zz;
      y = a*yy;
      z = fac*a*zz;
      cancel(x,y,z);
      return 1;
    }
  else // shouldn't happen as we tested solubility earlier
    {
      cout << "Problem in solve_conic (parameters (a,b,c,d)=("<<a<<","<<b<<","<<c<<","<<d<<"))\n";
      cout << "testlocsol() predicted solubility but no solution found!\n";
      x=0; y=0; z=0; return 0;
    }
}
  
int solve_conic_diag_nontriv(const bigint& a, const bigintArray& aplist,
		     const bigint& b, const bigintArray& bplist,
                bigint& x, bigint& y, bigint& z);
     // Solves xx-azz=byy for (x,y,z) not (0,0,0) and returns 1
     // or returns 0 if not possible
     // Should have a, b non-zero square-free, their prime divisors in aplist, bplist 
     // Here |b| >= |a|, |b| >=2, a!=1

int solve_conic_diag(const bigint& a, const bigintArray& aplist,
		     const bigint& b, const bigintArray& bplist,
                bigint& x, bigint& y, bigint& z)
     // Solves xx-azz=byy for (x,y,z) not (0,0,0) and returns 1
     // or returns 0 if not possible
     // Should have a, b non-zero square-free, their prime divisors in aplist, bplist 
     //
     // Here trivial cases are dealt with, non-trivial passed on
{
#ifdef DEBUG_CONIC
  cout << "In solve_conic_diag with a = " << a << ", b = " << b << endl;
#endif
  if(is_one(b)) 
    {
      x=1; y=1; z=0; 
#ifdef DEBUG_CONIC
      cout << "...returns ";  show_xyz(x,y,z);
#endif
      return 1;
    }
  if(is_one(a)) 
    {
      x=1; y=0; z=1; 
#ifdef DEBUG_CONIC
      cout << "...returns ";  show_xyz(x,y,z);
#endif
      return 1;
    }

  if(abs(a)>abs(b)) return solve_conic_diag(b,bplist,a,aplist,x,z,y);

  // Now |a|<=|b|, neither a nor b = 1

  if(is_one(-b)) 
    {
#ifdef DEBUG_CONIC
      cout << "...returns fail (a=b=-1)\n";
#endif
      return 0; // since a must be -1 too and xx+zz=-yy insoluble
    }

  if(b==-a) 
    {
      x=0; y=1; z=1; 
#ifdef DEBUG_CONIC
      cout << "...returns ";  show_xyz(x,y,z);
#endif
      return 1;
    }

  if(b==a) 
    {
      int res = solve_conic_diag(-1,bigintArray(0),a,aplist,y,x,z);
      x*=a;
      return res;
    }

  // Now |b| >= 2 , |b|>|a| and a is NOT square
  return solve_conic_diag_nontriv(a,aplist,b,bplist,x,y,z);
}

#if ((METHOD==2)||(METHOD==1)||(METHOD==0))

int solve_conic_diag_nontriv(const bigint& a, const bigintArray& aplist,
		     const bigint& b, const bigintArray& bplist,
                bigint& x, bigint& y, bigint& z)
     // Solves xx-azz=byy for (x,y,z) not (0,0,0) and returns 1
     // or returns 0 if not possible
     // Should have a, b non-zero square-free, their prime divisors in aplist, bplist 
     // Here |b| > |a|, |b| >=2, a!=1
{
  bigint x0, t, t0, t1, x1, y1, z1;  int res; bigintArray tplist;
  res = modsqrt(a,bplist,x0); // Solves x^2=a mod b, returns success/fail
  if(!res) 
    {
#ifdef DEBUG_CONIC
      cout << "...returns fail (no modular sqrt)\n";
#endif
      return 0; // since a must be -1 too and xx+zz=-yy insoluble
    }

  t=b;
  bigint newt = (sqr(x0)-a)/b;
  bigint m11=1, m21=0, m33=1, temp, fac;
#if (METHOD==0) // no reduction at all
  t=newt;
  m11=x0; m21=1; m33=t;
#else 
#if (METHOD==1) // algebraic reduction
#ifdef DEBUG_CONIC_2
  cout<<"t="<<t<<", newt="<<newt<<endl;
#endif
  while(abs(newt)<abs(t))
    {
      temp = x0*m11 +  a*m21;
      m21  =    m11 + x0*m21;
      m11  = temp;
      x0=mod(-x0,newt);
      t=newt;
      newt=(sqr(x0)-a)/t;
#ifdef DEBUG_CONIC_2
      cout<<"newt="<<newt<<endl;
#endif
      fac=gcd(gcd(m11,m21),m33);
      m11/=fac; m21/=fac; m33/=fac;
      m33=m33*t;
    }
#else // (METHOD==2): lattice reduction
#ifdef DEBUG_REDUCE
  cout<<"Before lattice reduction, x0="<<x0<<", t="<<newt<<endl;
  long count=0;
#endif
  bigint absa=abs(a);
  bigint u1=1, v1=0, u2=0, v2=1, w;
  bigint n1=absa+sqr(x0), dot=b*x0, n2;
  bigint alpha = roundover(dot, n1);  // nearest int to quotient
  int reduced = is_zero(alpha);
  if(!reduced)
    {
      u2-=alpha*u1;
      v2-=alpha*v1;
    }
  n2 = absa*sqr(u2) + sqr(u2*x0+v2*b);
  reduced=(n2>=n1);
#ifdef DEBUG_REDUCE
  //  cout<<"Entering lattice reduction loop, u1="<<u1<<", v1="<<v1<<", n1="<<n1<<endl;
  //  cout<<"Entering lattice reduction loop, u2="<<u2<<", v2="<<v2<<", n2="<<n2<<endl;
  //  cout<<"(alpha was "<<alpha<<")\n";
#endif
  while(!reduced)
    {
      swap(u1,u2);  // w=u1; u1=u2; u2=w;
      swap(v1,v2);  // w=v1; v1=v2; v2=w;
      swap(n1,n2);  // w=n1; n1=n2; n2=w;
      dot=absa*u1*u2+(u1*x0+v1*b)*(u2*x0+v2*b);
      alpha=roundover(dot,n1);
      reduced = is_zero(alpha);
      if(!reduced)
	{
	  u2-=alpha*u1;
	  v2-=alpha*v1;
	  n2 = absa*sqr(u2) + sqr(u2*x0+v2*b);
	}
      reduced=(n2>=n1);
#ifdef DEBUG_REDUCE
      count++;
      //      cout<<"In lattice reduction loop, u1="<<u1<<", v1="<<v1<<", n1="<<n1<<endl;
      //      cout<<"In lattice reduction loop, u2="<<u2<<", v2="<<v2<<", n2="<<n2<<endl;
#endif
    }
  m11 = u1*x0+v1*b;  
  m21 = u1;
  t   = (sqr(m11)-a*sqr(m21))/b;
  m33 = t;
#ifdef DEBUG_REDUCE
  cout<<"("<<count<<" steps in reduction)\n";
  cout<<"After  lattice reduction, x0="<<m11<<", z0="<<m21<<", t="<<t<<endl;
#endif

#endif
#endif
  //Now use recursion unless t=square already
  if(is_one(t))
    {
      x=m11; y=1; z=m21;
#ifdef DEBUG_CONIC
      cout << "...returns ";  show_xyz(x,y,z);
#endif
      return 1;
    }
  sqfdecomp(t,t0,t1,tplist);
  if(is_one(t0))
    {
      x=m11; y=t1; z=m21;
#ifdef DEBUG_CONIC
      cout << "...returns ";  show_xyz(x,y,z);
#endif
      return 1;
    }
  m33/=t1;

  res = solve_conic_diag(a,aplist,t0,tplist,x1,y1,z1);
  if(!res) 
    {
#ifdef DEBUG_CONIC
      cout << "...returns fail\n";
#endif
      return 0; // since a must be -1 too and xx+zz=-yy insoluble
    }
  x =  m11*x1 + a*m21*z1;
  z =  m21*x1 +   m11*z1;
  y =  m33*y1;

  bigint g = cancel1(x,y,z);
#ifdef DEBUG_CONIC
  cout << "...returns (after cancelling "<<g<<") ";  show_xyz(x,y,z);
#endif
  return 1;
}

#else
#if (METHOD==3) // JC's  version 11/3/97: no good

int solve_conic_diag_nontriv(const bigint& a, const bigintArray& aplist,
		     const bigint& b, const bigintArray& bplist,
                bigint& x, bigint& y, bigint& z)
     // Solves xx-azz=byy for (x,y,z) not (0,0,0) and returns 1
     // or returns 0 if not possible
     // Should have a, b non-zero square-free, their prime divisors in aplist, bplist 
     // Here |b| >= |a|, |b| >=2, a!=1
{
  bigint x0, t, t0, t1, x1, y1, z1;  int res; bigintArray tplist;
  res = modsqrt(a,bplist,x0); // Solves x^2=a mod b, returns success/fail
  if(!res) 
    {
#ifdef DEBUG_CONIC
      cout << "...returns fail (no modular sqrt)\n";
#endif
      return 0; // since a must be -1 too and xx+zz=-yy insoluble
    }

  // Always have v ^2 - a*u ^2  = b*c
  //             vv^2 - a*uu^2  = c*newc

  bigint u=1, v=x0, temp, c, uu, vv, nor, newc;
  c = (x0*x0-a)/b;
  vv = mod(-x0,c), uu=1;
  nor = vv*vv-a*uu*uu;
  //  if(!div(c,nor)) cout<<"Error: "<<nor<<" not divisible by c\n";
  newc = nor/c;
#ifdef DEBUG_CONIC_2
  cout<<"   c="<<c<<"\n";
  cout<<"newc="<<newc<<"\n";
  //  cout<<"(v,u)=("<<v<<","<<u<<")\n";
#endif
  while(!is_zero(newc)&&(abs(newc)<abs(c)))  // update u, v
    {
      temp = (u*vv+v*uu)/c;    // exact by defn of uu, vv
      v    = (v*vv+a*u*uu)/c;  // exact since v^2-au^2=0(mod c)
      u    = temp;
      //      cout<<"(v,u)=("<<v<<","<<u<<")\n";
      uu = mod(u,abs(newc));
      vv = mod(-v,abs(newc));
      //      cout<<"(vv,uu)=("<<vv<<","<<uu<<")\n";
      c    = newc;
      nor = vv*vv-a*uu*uu;
      //      if(nor==0) cout<<"Error: nor=0 when uu="<<uu<<", vv="<<vv<<endl;
      //      if(!div(c,nor)) cout<<"Error: "<<nor<<" not divisible by c\n";
      newc = nor/c;
#ifdef DEBUG_CONIC_2
  cout<<"newc="<<newc<<"\n";
#endif
    }
  // now we have v^2-a*u^2=b*c with smallest c, and
  // use recursion to solve x^2-a*z^2=c*y^2

  sqfdecomp(c,t0,t1,tplist);
  res = solve_conic_diag(a,aplist,t0,tplist,x1,y1,z1);
  if(!res) 
    {
#ifdef DEBUG_CONIC
      cout << "...returns fail\n";
#endif
      return 0; // since a must be -1 too and xx+zz=-yy insoluble
    }
  x =  v*x1 + a*u*z1;
  z =  u*x1 +   v*z1;
  y =  t0*t1*y1;

  bigint g = cancel1(x,y,z);
#ifdef DEBUG_CONIC
  cout << "...returns (after cancelling "<<g<<") ";  show_xyz(x,y,z);
#endif
  return 1;
}

#endif
#endif

int testsol(const bigint& a, const bigint& b, const bigint& c, const bigint& d,
                const bigint& x, const bigint& y, const bigint& z, int verb)
{
  int triv = is_zero(x) && is_zero(y) && is_zero(z);
  if(triv)
    {
      if(verb) cout << "Trivial solution!\n";
      return 0;
    }
  bigint t = a*x*x+b*x*z+c*z*z-d*y*y;
  if(is_zero(t))
    {
      if(verb) cout << "Solution OK!\n";
      return 1;
    }
  else 
    {
      if(verb) cout << "Solution wrong!\n";
      return 0;
    }

}

int testlocsol(const bigint& a, 
	       const bigint& b, 
	       const bigint& c)
// tests if ax^2+by^2+cz^2=0 is soluble, where a, b, c are pairwise
// coprime and square-free
{
  bigintArray alist=pdivs(a);
  bigintArray blist=pdivs(b);
  bigintArray clist=pdivs(c);
  return testlocsol(a,alist,b,blist,c,clist);
} 

int testlocsol(const bigint& a, const bigintArray& alist, 
	       const bigint& b, const bigintArray& blist, 
	       const bigint& c, const bigintArray& clist) 
// tests if ax^2+by^2+cz^2=0 is soluble, where a, b, c are pairwise
// coprime and square-free, their prime factors being in alist etc.
{
  int as=sign(a), bs=sign(b), cs=sign(c);
  if((as==bs)&&(bs==cs)) 
    {
      //      cout<<"testlocsol("<<a<<","<<b<<","<<c<<") returning 0 because of signs\n";
      return 0;
    }  
  bigint p, two(2); long ip;
  bigint mab=-a*b; 
  for(ip=0; ip<clist.length; ip++) 
    {
      p=clist(ip);
      if(p==two) continue;
      if(legendre(mab,p)!=1) 
	{
// 	  cout<<"testlocsol fails legendre(mab,p) with "
// 	      <<"(a,b,p)=("<<a<<","<<b<<","<<p<<")\n"; 
	  return 0;
	}
    }
  bigint mbc=-b*c; 
  for(ip=0; ip<alist.length; ip++) 
    {
      p=alist(ip);
      if(p==two) continue;
      if(legendre(mbc,p)!=1)
	{
// 	  cout<<"testlocsol fails legendre(mbc,p) with "
// 	      <<"(b,c,p)=("<<b<<","<<c<<","<<p<<")\n"; 
	  return 0;
	}
    }
  bigint mca=-c*a;
  for(ip=0; ip<blist.length; ip++) 
    {
      p=blist(ip);
      if(p==two) continue;
      if(legendre(mca,p)!=1)
	{
// 	  cout<<"testlocsol fails legendre(mca,p) with "
// 	      <<"(c,a,p)=("<<c<<","<<a<<","<<p<<")\n"; 
	  return 0;
	}
    }
  return 1;
} 

int testlocsol(const bigint& a, const bigintArray& alist, 
	       const bigint& b, const bigintArray& blist)
// tests if ax^2+by^2=z^2 is soluble, where a, b are
// square-free, their prime factors being in alist and blist.
{
  // Avoid any factorization and gcd computation using the primes given
  bigint p, a0(1), b0(1), c(-1);
  long ip, ia=0, ib=0, ic=0;
  bigintArray a0list(alist.length), b0list(blist.length), clist(alist.length);

  long sa=sign(a), sb=sign(b);
  if((sa<0)&&(sb<0)) 
    {
//       cout<<"testlocsol("<<a<<","<<b<<") returning 0 because of signs\n";
      return 0;  // nothing more to do as no real solution
    }
  if(sa<0) a0.negate();
  if(sb<0) b0.negate();
  
  for(ip=0; ip<alist.length; ip++)
    {
      p=alist(ip);
      if(div(p,b)) 
	{
	  c*=p; clist[ic++]=p;
	}
      else
	{
	  a0*=p; a0list[ia++]=p;
	}
    }
  for(ip=0; ip<blist.length; ip++)
    {
      p=blist(ip);
      if(!div(p,c)) {b0*=p; b0list[ib++]=p;}
    }  
  a0list.truncate(ia);
  b0list.truncate(ib);
  clist.truncate(ic);
  
#if(1)
  if((a!=-a0*c)||(b!=-b0*c)||(abs(c)!=gcd(a,b)))
    {
      cout<<"Error: (a,b)=("<<a<<","<<b<<") gives a0="<<a0<<", b0="<<b0<<", c="<<c<<endl;
    }
  //  cout<<"Calling testlocsol(a,b,c) with a,b,c="<<a0<<","<<b0<<","<<c<<"\n";
  //  cout<<"alist="<<a0list<<endl;
  //  cout<<"blist="<<b0list<<endl;
  //  cout<<"clist="<<clist<<endl;
#endif

  // Now a=a0*c, b=b0*c, c=gcd(a,b), and a0list, b0list, clist hold their primes
  
  return testlocsol(a0,a0list,b0,b0list,c,clist);

}
