//
// cperiods.cc -- periods and lattices
//

#include "cperiods.h"

//#define DEBUG 1

//
// periods constructed from a curve, normalized to lattice form
//

// uses the following functions:

#ifndef LiDIA
void swap(bigcomplex& a, bigcomplex& b)
{
  bigcomplex c(a); a=b; b=c;
}
#endif

// Reorders 3 complex nos so real parts are decreasing
void reorder1(bigcomplex& a, bigcomplex& b, bigcomplex& c)
{
  if (real(a) < real(c)) swap(a,c);
  if (real(a) < real(b)) swap(a,b);
  else if (real(b) < real(c)) swap(b,c);
}
 
//reorders 3 complex nos so e1 is real if any (
void reorder2(bigcomplex& e1, bigcomplex& e2, bigcomplex& e3)
{
  if (is_real(e1)) return;
  else if (is_real(e2)) {swap(e1,e2); return;}
  else if (is_real(e3)) {swap(e1,e3); return;}
}
 
//Computes periods of a curve given the 3 2-division points (i.e. the three
//roots of the cubic)
void eiperiods(bigcomplex e1, bigcomplex e2, bigcomplex e3, 
               bigcomplex& w1, bigcomplex& w2)
{
  bigcomplex a(sqrt(e3-e1)); 
  bigcomplex b(sqrt(e3-e2)); 
  bigcomplex c(sqrt(e2-e1)); 
#ifdef DEBUG
  cout<<"In eiperiods with a = " << a << ", b = "<<b<< ", c = " << c << endl;
#endif
  bigcomplex agm1 = cagm(a,b);
  bigcomplex agm2 = cagm(a,c);
#ifdef DEBUG
  cout<<"agm1=agm(a,b)="<<agm1<<endl;
  cout<<"agm2=agm(a,c)="<<agm2<<endl;
#endif
  bigfloat pi = Pi();
  w1= bigcomplex(pi,0)/agm1;
  w2= bigcomplex(0,pi)/agm2;
#ifdef DEBUG
  cout<<"Leaving eiperiods with w1 = " << w1 << ", w2 = "<<w2 << endl;    
#endif
}

// Gets the 3 2-division points given the coefficients 
void getei(bigcomplex a1,bigcomplex a2,bigcomplex a3,bigcomplex a4,bigcomplex a6,
	   bigcomplex& e1, bigcomplex& e2, bigcomplex& e3)
{
  bigcomplex c1 = a2 + a1*a1/4;
  bigcomplex c2 = a4 + a1*a3/2;
  bigcomplex c3 = a6 + a3*a3/4;
  bigcomplex* ei = solvecubic(c1,c2,c3);
  e1 = ei[0];  e2 = ei[1];  e3 = ei[2];
  delete [] ei;
}

Cperiods::Cperiods(const Curve& E)
{
  bigint a1,a2,a3,a4,a6;    
  E.getai(a1,a2,a3,a4,a6);

  bigcomplex ca1=I2bigfloat(a1);
  bigcomplex ca2=I2bigfloat(a2);
  bigcomplex ca3=I2bigfloat(a3);
  bigcomplex ca4=I2bigfloat(a4);
  bigcomplex ca6=I2bigfloat(a6);
  
  getei(ca1,ca2,ca3,ca4,ca6,e1,e2,e3);
  
  int allrealroots = is_real(e1) && is_real(e2) && is_real(e3);
#ifdef DEBUG
  cout << "original e1:" << e1 << "\ne2:" << e2 << "\ne3:" << e3 << "\n";
  if(!allrealroots) cout << "NOT "; cout << "all real"<<endl;
#endif
  if (allrealroots) reorder1(e3,e2,e1); 
  else reorder2(e3,e2,e1);
// this ordering ensures that eiperiods will give w1,w2 per lattice norm:
// if allrealroots then w1 wil be real, w2 pure imag;
// if not, w1 will be real but w2 is not specially determined.
#ifdef DEBUG
  cout << "final e1:" << e1 << "\ne2:" << e2 << "\ne3:" << e3 << "\n";
#endif
  
  eiperiods(e1,e2,e3,w1,w2);
#ifdef DEBUG
  cout << "w1:" << w1 << " (should be real)\n";
  cout << "w2:" << w2 << " (real part should be 0 if all roots real, else half w1)\n";
  cout << "tau: " << (w2/w1) << "\n";
#endif  
  w1 = bigcomplex( real(w1), 0 );  // force to be real as it ought to be anyway
  tau = w2/w1;
#ifdef DEBUG
  cout << "After fixing (1) -- \nw1:" << w1 << " (should be real)\n";
  cout << "w2:" << w2 << " (real part should be 0 if all roots real, else half w1)\n";
  cout << "tau: " << tau << ", real(tau) = "<<real(tau)<<"\n";
#endif  
  if(real(tau)<-0.1) {tau+=1; w2+=w1;}
  norm_code = 1+allrealroots;  //ie type 1 or 2 as appropriate
#ifdef DEBUG
  cout << "After fixing (2) -- \nw1:" << w1 << "\nw2:" << w2 << "\ntau:" << tau << "\n";
#endif  
}


void Cperiods::norm_lattice()
{
  cerr <<" ## norm_lattice() not implemented yet\n";
}

// NB In compproc.h/cc's normalize and getc4c6 the normalization is such that
//    tau = w1/w2 in fundamental region; here we have tau=w2/w1 (better for 
//    lattice normalization) so must call them with w1, w2 reversed.

void Cperiods::norm_region()
{
   if (norm_code!=3)
     {
       tau = normalize(w2,w1);  // NB reverse params
       norm_code = 3;
       qtau = q(tau);
       if(abs(qtau)>0.99)
	 {
	   cout << "Warning from Cperiods::norm_region: qtau = " << qtau << " is not small!\n";
	 }
       w1squared = w1*w1;
       w1cubed   = w1*w1squared;
       bigcomplex term = bigfloat(1), qtm = qtau;
       sum3=0;
       for (bigfloat m=1; ! is_approx_zero(term); m+=1)
	 { term = qtm*m / (1 - qtm);
	   qtm *= qtau;
	   sum3 += term;
#ifdef DEBUG
	   cout<<"term = "<<term<<", sum3 = "<<sum3<<endl;      
#endif
	 }
#ifdef DEBUG
       cout<<"final sum3 = "<<sum3<<endl;      
#endif
       sum3 = bigfloat(1)/12 - 2*sum3;
     }
}

bigcomplex Cperiods::X_coord(const bigcomplex& qz)
{
  bigcomplex sum(sum3), term(1),  qtm(1), w;
  while ( ! is_approx_zero(term) )
    { w = qtm*qz;
      term = w / pow((1 - w), 2);
      qtm *= qtau;
      sum += term;
#ifdef DEBUG_XY
      cout<<"X-term = "<<term<<", sum = "<<sum<<endl;      
#endif
    }
  term = 1; qtm = qtau;
  while ( ! is_approx_zero(term) )
    { w = qtm / qz;
      qtm *= qtau;
      term = w / pow((1 - w), 2);
      sum += term;
#ifdef DEBUG_XY
      cout<<"X-term = "<<term<<", sum = "<<sum<<endl;      
#endif
    }
  bigcomplex ans = sum*TWOPIEYE*TWOPIEYE;
#ifdef DEBUG_XY
  cout<<"X_coord returning ans = "<<ans<<endl;
#endif
  return ans;
}

bigcomplex Cperiods::Y_coord(const bigcomplex& qz)
{
  bigcomplex sum(0), term(1), qtm(1), w;
  while ( !is_approx_zero(term) )
    { w = qtm*qz;
      term = w*(1 + w) / pow((1 - w), 3);
      qtm *= qtau;
      sum += term;
#ifdef DEBUG_XY
      cout<<"Y-term = "<<term<<", sum = "<<sum<<endl;      
#endif
    }
  qtm = qtau; term = 1;
  while ( !is_approx_zero(term) )
    { w = qtm / qz;
      term = w*(1 + w) / pow((w - 1), 3);
      qtm *= qtau;
      sum += term;
#ifdef DEBUG_XY
      cout<<"Y-term = "<<term<<", sum = "<<sum<<endl;      
#endif
    }
  bigcomplex ans =  sum * TWOPIEYE * TWOPIEYE * TWOPIEYE;
#ifdef DEBUG_XY
  cout<<"Y_coord returning ans = "<<ans<<endl;
#endif
  return ans;
}


void Cperiods::XY_coords(bigcomplex& X, bigcomplex& Y, const bigcomplex& z)
{
  norm_region();
  bigcomplex z1 = z/w1;
  bigcomplex qz = q(z1);
  while(abs(qz)>0.9) qz*=qtau;
#ifdef DEBUG_XY
  cout<<"In XY-coords with qtau = " << qtau << ", qz = " << qz << endl;
#endif
  X = X_coord(qz) / w1squared;
  Y = Y_coord(qz) / w1cubed;
  return;
}
  
void makec4c6(Cperiods& cp, bigcomplex& cc4, bigcomplex& cc6)
{
  cp.norm_region();  // normed for tau in fundamental region
  bigcomplex w1, w2;
  cp.getwi(w1, w2);
#ifdef DEBUG_XY
  cout << "About to call getc4c6() with first w2 = " << w2 << "\n";
  cout << "second w1 = " << w1 << "\n";
  cout << "tau = w2/w1 = " << w2/w1 << "\n";
#endif
  getc4c6(w2,w1,cc4,cc6);  // from compproc.h
#ifdef DEBUG_XY
  cout << "Done, returned values c4 = " << cc4 << ", c6 = " << cc6 << "\n";
#endif
}

Curve trans_to_curve(Cperiods& cp)
{
  bigint c4, c6;
  bigcomplex cc4, cc6;
  bigfloat rc4, rc6;

  makec4c6(cp, cc4, cc6);
  if ( abs(imag(cc4)) > 1.0e-10 )
    cerr<<" ## trans_to_curve: possible error taking "<<cc4<<" as real"<<endl;
  if ( abs(imag(cc6)) > 1.0e-10 )
    cerr<<" ## trans_to_curve: possible error taking "<<cc6<<" as real"<<endl;

  rc4 = real(cc4);
  rc6 = real(cc6);

  c4 = Iround( rc4 );
  c6 = Iround( rc6 );

//  if ( abs(rc4 - I2bigfloat(c4)) > 1.0e-5 )
//    cerr<<" ## trans_to_curve: possible error taking "<<rc4<<" as integer"
//      <<endl;
//  if ( abs(rc6 - I2bigfloat(c6)) > 1.0e-5 )
//    cerr<<" ## trans_to_curve: possible error taking "<<rc6<<" as integer"
//      <<endl;

  return Curve(c4,c6);
}

// end of file cperiods.cc
