// DESC2.CC  Second descent procedure for use with descent via 2-isogeny

#include "mquartic.h"
#include "conic.h"
#include "mlocsol.h"
#include "mglobsol.h"
#include "transform.h"
#include "minim.h"
#include "reduce.h"
#include "sqfdiv.h"
#include "desc2.h"

int desc2(const bigint& c, const bigint& d1, const bigint& d2,
	  const bigintArray& plist, const bigintArray& supp, const bigintArray& bgens,
	  long mask,  double hlim,
	  bigint& x, bigint& y, bigint& z, int verb, int alldesc)
// Works on homogeneous space (d1,0,c,0,d2) (assumed ELS)
// Returns 
//   -1 if it certainly has no points (if no ELS descendents)
//   +1 if it has a point (coordinates returned in x, y, z)
//    0 if undecided (ELS descendents exist but no rational points were found)
//
//  plist is a complete list of bad primes (including 2)
//  supp is a list of primes (and -1) dividing d'
//  bgens is a list of generators of the "opposite" Selmer group B', a subgroup
//        of group A' = the span of supp
//  mask enables one to loop over A' modulo B'
//
// if alldesc==1 it does not stop when it finds one descendent with a point on it,
// but goes on to look at all the others.
{
  int xverb = (verb>1);
  int els_descendent_exists=0;
  bigint zero(0), one(1);
  bigint x0,y0,z0,d0, p, resultant;
  int res;
  bigint m00(1), m11(1), m12(0), m21(0), m22(1);
  bigint q10,q11,q12,q30,q31,q32;
  long i;

  bigint d = d1*d2, cdash = -2*c, ddash = sqr(c)-4*d;
  int add2tosupp = (supp.locate(2)==-1)&&(val(2,ddash)==4);  
// For we are on E' and the original d was odd

  // Step 1: solve the conic d1*x0^2 + c*x0*z0 + d2*z0^2 = y0^2 with y0!=0

  int ddash_is_square = isqrt(ddash,d0);
  if(ddash_is_square) // special case to make sure y0!=0
    {
      if(xverb) cout<<"d' is square\n";
      x0 = (1+d1)*d0-c*(1-d1);
      y0 = 2*d1*d0;
      z0 = 2*d1*(1-d1);
      cancel(x0,y0,z0);
    }
  else
    {
      res = solve_conic(d1,c,d2,1,x0,y0,z0);
      if(!res) {cout<<"solve_conic failed\n"; return -1;}
    }

  // Step 2: Parametrize all solutions of that conic

  q10 = x0;
  q11 = 2*(c*x0+2*d2*z0);
  q12 = ddash*x0;
  if(xverb) cout<<"original q1-coeffs: "<<q10<<", "<<q11<<", "<<q12<<endl;
  q30 = z0;
  q31 = -2*(2*d1*x0+c*z0);
  q32 = ddash*z0;
  if(xverb) cout<<"original q3-coeffs: "<<q30<<", "<<q31<<", "<<q32<<endl;
  

  // These quadratics have discriminants 16*d_2*y0^2, 16*d1*y0^2,
  // and resultant 16*d'*y0^4.  First we adjust them to remove the y0.

  bigint pf = abs(y0); bigint pf2=sqr(pf);
  
  if(pf>1) // else nothing to do here
    {
      // Factor pf into parts coprime to x0, z0:
      bigint g=1, x0pow=x0;
      bigint pfb= gcd(pf,x0);
      while(pfb>g)
	{
	  x0pow*=x0;
	  g=pfb;
	  pfb=gcd(pf,x0pow);
	}
      bigint pfa=pf/pfb;

      // Now pfa is coprime to x0 and pfb contains only primes dividing x0 
      // which therefore do not divide z0 sinve gcd(x0,y0,z0)=1

      bigint pfa2=sqr(pfa);
      bigint x0inv = invmod(x0,pfa2);
      bigint e = mod(-c-2*d2*z0*x0inv,pfa2);

      if(pfb>1)
	{
	  bigint pfb2=sqr(pfb);
	  bigint z0inv = invmod(z0,pfb2);
	  bigint eb = mod(c+2*d1*x0*z0inv,pfb2);
	  bigint u, v;
	  bigint g = bezout(pfa2,pfb2,u,v); // = 1
	  e = mod((e*v*pfb2+eb*u*pfa2),pf2);
	}

      if(xverb) cout<<"Shifting by e = " << e << endl;
      
      q12+=e*(q11+e*q10);
      q11+=2*e*q10;
      if(!divide_exact(q11,pf,q11)) cout<<"Problem: pf ndiv q11 = "<<q11<<endl;
      if(!divide_exact(q12,pf2,q12)) cout<<"Problem: pf ndiv q12 = "<<q12<<endl;
      if(xverb) cout<<"reduced q1-coeffs: "<<q10<<", "<<q11<<", "<<q12<<endl;
      
      q32+=e*(q31+e*q30);
      q31+=2*e*q30;
      if(!divide_exact(q31,pf,q31)) cout<<"Problem: pf ndiv q31 = "<<q31<<endl;
      if(!divide_exact(q32,pf2,q32)) cout<<"Problem: pf ndiv q32 = "<<q32<<endl;
      if(xverb) cout<<"reduced q3-coeffs: "<<q30<<", "<<q31<<", "<<q32<<endl;
    }
      
  if(xverb) 
    {
      resultant = sqr(q30)*sqr(q12) + sqr(q10)*sqr(q32) + q32*q30*sqr(q11)
	- q31*q11*(q30*q12 + q10*q32) + (sqr(q31)-2*q30*q32)*q10*q12 ;
      cout<<"reduced q1-coeffs: "<<q10<<", "<<q11<<", "<<q12<<endl;
      cout<<"reduced q3-coeffs: "<<q30<<", "<<q31<<", "<<q32<<endl;
      cout<<"Current resultant   = "<<resultant<<endl;
      cout<<"resultant/d' (=16?) = "<<(resultant/ddash)<<endl;
    }

#if(0)  // This does not work properly, but is possibly not worth it anyway.
  // Final reduction to remove the 16: each has a multiple root mod 2, 
  // and they are the same

  // First find the multiple root:

  bigint qt0=q30, qt1=q31, qt2=q32;
  while(even(qt0)&&even(qt1)&&even(qt2)) {qt0/=2; qt1/=2; qt2/=2;}

  if(even(qt0))  // multiple root mod 2 is at infinity
    {
      p=q12; divide_exact(q10,4,q12); divide_exact(q11,2,q11); q10=p;
      p=q32; divide_exact(q30,4,q32); divide_exact(q31,2,q31); q30=p;
    }
  else if(even(qt2))  // multiple root mod 2 is at 0
    {
      divide_exact(q12,4,q12); divide_exact(q11,2,q11);
      divide_exact(q32,4,q32); divide_exact(q31,2,q31);
    }
  else // multiple root mod 2 is at 1
    {
      q12 += (q10+q11);  divide_exact(q12,4,q12); 
      divide_exact(q11,2,q11); q11 += q10;

      q32 += (q30+q31);  divide_exact(q32,4,q32); 
      divide_exact(q31,2,q31); q31 += q30;
    }
#endif
#if(0)
  // Have a look at intermediate quartic q1*q3=square
  // Minimize it and hence reduce q1, q3 further if possible
  
  bigint ta=q10*q30;
  bigint tb=q10*q31+q11*q30;
  bigint tc=q10*q32+q11*q31+q12*q30;
  bigint td=        q11*q32+q12*q31;
  bigint te=                q12*q32;
  bigint tcont = g_content(ta,tb,tc,td,te);
		  
  if(tcont>1)
    {
      tcont=sqr(tcont);
      ta/=tcont; tb/=tcont; tc/=tcont; td/=tcont; te/=tcont;
    }
  bigint tI = II(ta,tb,tc,td,te);
  bigint tJ = JJ(ta,tb,tc,td,te);
  bigint tdisc = 4*pow(tI,3)-sqr(tJ);
  bigintArray tgbadp = pdivs(tdisc);

  minim_all(ta,tb,tc,td,te,tI,tJ,tgbadp,m00,m11,m12,m21,m22,1,0);
  reduce(ta,tb,tc,td,te,m00,m11,m12,m21,m22);
  tdisc = 4*pow(tI,3)-sqr(tJ);
  
  if(abs(m00)>1)
    {
      if(xverb) cout<<"Further reduction of q1 and q3:\n";
      bigint t10 =   q10*sqr(m11) + q11*m11*m21           +   q12*sqr(m21);
      bigint t11 = 2*q10*m11*m12  + q11*(m11*m22+m12*m21) + 2*q12*m21*m22;
      bigint t12 =   q10*sqr(m12) + q11*m12*m22           +   q12*sqr(m22);
      bigint gcdut=gcd(t10,t11);
      gcdut=gcd(gcdut,t12);
      bigint u10 =   q30*sqr(m11) + q31*m11*m21           +   q32*sqr(m21);
      bigint u11 = 2*q30*m11*m12  + q31*(m11*m22+m12*m21) + 2*q32*m21*m22;
      bigint u12 =   q30*sqr(m12) + q31*m12*m22           +   q32*sqr(m22);
      gcdut=gcd(gcdut,u10);
      gcdut=gcd(gcdut,u11);
      gcdut=gcd(gcdut,u12);
      if(xverb)
	if(gcdut!=m00)
	  cout<<"In simultaneous quadratic reduction, m00="<<m00
	      <<", but common factor = "<<gcdut<<"\n";
      divide_exact(t10,gcdut,q10);
      divide_exact(t11,gcdut,q11);
      divide_exact(t12,gcdut,q12);
      divide_exact(u10,gcdut,q30);
      divide_exact(u11,gcdut,q31);
      divide_exact(u12,gcdut,q32);
      if(xverb) 
	{
	  cout<<"further reduction factor = "<<gcdut<<endl;
	  cout<<"further reduced q1-coeffs: "<<q10<<", "<<q11<<", "<<q12<<endl;
	  cout<<"further reduced q3-coeffs: "<<q30<<", "<<q31<<", "<<q32<<endl;
	}
    }
#endif
  
  if(xverb) 
    {
      cout<<"reduced q1-coeffs: "<<q10<<", "<<q11<<", "<<q12<<endl;
      cout<<"reduced q3-coeffs: "<<q30<<", "<<q31<<", "<<q32<<endl;
    }

  // Step 3

  if(xverb) 
    {
      resultant = sqr(q30)*sqr(q12) + sqr(q10)*sqr(q32) + q32*q30*sqr(q11)
	- q31*q11*(q30*q12 + q10*q32) + (sqr(q31)-2*q30*q32)*q10*q12 ;
      cout<<"resultant = " <<resultant<<endl;
      cout<<"resultant/d'      = "<<(resultant/ddash)<<endl;
      cout<<"support of d': "<<supp<<endl;
    }
  long nsupp = supp.length;
  long nd3=1<<nsupp, id3, jd3=0, index;

  // Step 4: For all square-free divisors of the resultant, attempt to
  //         form a descendent quartic

  bigint d3, keepd3, s0, t0, u0, s1, t1, u1;
  int looking=1, found=0;

  for(id3=0; (id3<nd3)&&(looking||alldesc); id3++) // must start at 0
    {
      d3=makenum(supp,id3);
      if(id3&mask) continue;

      int res = process_d3(d1,c,d2,d3,plist,hlim,q10,q11,q12,q30,q31,q32,x,y,z,verb);
      if(xverb) cout<<"process_d3("<<d3<<") returns "<<res<<endl;
      if(res!=-1)  // descendent is els, lift to S^(2) exists
	{
	  looking=0; 
	  keepd3=d3;
	}
      if(res==1)   // descendent is gls, lift to E/2E exists
	{
	  found=1;
	}

      if((looking||alldesc)&&add2tosupp)
	{
	  d3=2*d3;
	  res = process_d3(d1,c,d2,d3,plist,hlim,q10,q11,q12,q30,q31,q32,x,y,z,verb);
	  if(xverb) cout<<"process_d3("<<d3<<") returns "<<res<<endl;
	  if(res!=-1)  // descendent is els, lift to S^(2) exists
	    {
	      looking=0; 
	      keepd3=d3;
	    }
	  if(res==1)   // descendent is gls, lift to E/2E exists 
	    {
	      found=1;
	    }
	}
    }
  if(found) 
    {
      if(verb) 
	cout<<"Found a descendent with a rational point so terminating second descent step.\n";
      return +1;   // Found an els descendent with a point on it
    }
  if(looking) return -1; // No els descendents exist
  //
  // Go on searching on other descendents...
  //
  if(verb) cout<<"Found an els descendent but no rational point on it...\n";

  nd3=1<<bgens.length;
  long d3step=(2-ddash_is_square);
  long ndesc=nd3/d3step;
  if(nd3==d3step)
    {
      if(verb) cout<<"No further descendents.\n";
      return 0;
    }
  if(verb) 
    cout<<"We now construct and search the "<<(ndesc-1)<<" other descendents...\n";
  for(id3=d3step; (id3<nd3); id3+=d3step)
    {
      d3 = sqfmul(keepd3,makenum(bgens,id3));
      int res = process_d3(d1,c,d2,d3,plist,hlim,q10,q11,q12,q30,q31,q32,x,y,z,verb);
      if(xverb) cout<<"process_d3("<<d3<<") returns "<<res<<endl;
      if(res==1)   // descendent is gls, lift to E/2E exists
	{
	  found=1;
	  if(verb) 
	    cout<<"Found a descendent with a rational point so terminating second descent step.\n";
	  if(!alldesc) return +1;
	}
    }
  if(verb&&(!found)) cout<<"No rational point found on any of the other descendents...\n";
  return found;
  //
  //
  //
}  // end of desc2()


int process_d3(const bigint& d1, const bigint& c, const bigint& d2, 
	       const bigint& d3, const bigintArray& plist, double hlim,
	       const bigint& q10, const bigint& q11, const bigint& q12, 
	       const bigint& q30, const bigint& q31, const bigint& q32, 
	       bigint& x, bigint& y, bigint& z, int verb)
// Processes an individual d3
// Returns -1 if not els
//          0 if els but no point found
//         +1 if els and point found
{
  int xverb=verb>1;
  if(verb) cout<<"Processing d3 = "<<d3<<": \t";
  //  if(xverb) cout<<"\nplist = "<<plist<<"\n";
  bigint s0,u0,t0,s1,u1,t1,Q10,Q11,Q12,Q30,Q31,Q32,ga,gb,gc,gd,ge,cont;
  bigint p,ggI,ggJ,ggD,m00,m11,m12,m21,m22;
  bigint one(1), zero(0);
  bigintArray ggbadp, ggextrap;

  int resd3 = solve_conic(q10,q11,q12,d3,s0,u0,t0);
  if(resd3)
    {
      if(verb) cout<<"q1=d3 is soluble.\t";
      resd3 = solve_conic(q30,q31,q32,d3,s1,u1,t1);
      if(resd3)
	{
	  if(verb) cout<<"q1=d3 and q3=d3 both soluble, forming quartic.\n";
	      
	  Q10=s0*q12, Q11=2*t0*q12, Q12=-q10*s0-q11*t0;
	  Q30=-s0*q11-t0*q12, Q31=2*s0*q10, Q32=t0*q10;
	      
	  ga = q30*sqr(Q10) + q31*Q10*Q30 + q32*sqr(Q30);
	  
	  gb = q30*(2*Q10*Q11) + q31*(Q10*Q31+Q11*Q30) + q32*(2*Q30*Q31);
	  
	  gc = q30*(sqr(Q11) + 2*Q10*Q12) 
	    + q31*(Q10*Q32 + Q11*Q31 + Q12*Q30) 
	    + q32*(sqr(Q31) + 2*Q30*Q32);
	  
	  gd = q30*(2*Q11*Q12) + q31*(Q11*Q32 + Q12*Q31) + q32*(2*Q31*Q32);
	  
	  ge = q30*(sqr(Q12)) + q31*(Q12*Q32) + q32*(sqr(Q32));
	  
	  ga*=d3; gb*=d3; gc*=d3; gd*=d3; ge*=d3;
	  
	  // Simplify quartic by dividing by square-part of content
	  cont = g_content(ga,gb,gc,gd,ge);
	  
	  if(xverb) cout<<"Dividing quartic by "<<cont<<" squared\n";
	  cont=sqr(cont);
	  ga/=cont; gb/=cont; gc/=cont; gd/=cont; ge/=cont;
	  
	  quartic gg(ga,gb,gc,gd,ge);
	  ggI = gg.getI();
	  ggJ = gg.getJ();
	  ggD = gg.getdisc();
	  ggbadp = bigintArray(100);
	  long ip, nbadp=plist.length;
	  for(ip=0; ip<nbadp; ip++) 
	    {
	      ggbadp[ip]=plist(ip);
	      divide_out(ggD,plist(ip));
	    }
	  ggextrap = pdivs(ggD);
	  for(ip=0; ip<ggextrap.length; ip++)
	    {
	      ggbadp[nbadp++] = ggextrap(ip);
	    }
	  ggbadp.truncate(nbadp);
	  if(xverb)
	    {
	      cout<<"Quartic gg is " << gg << endl;
	      cout<<"I(gg) = "<<ggI<<endl;
	      cout<<"J(gg) = "<<ggJ<<endl;
	      cout<<"extra bad primes = "<<ggextrap<<endl;
	      cout<<"bad primes = "<<ggbadp<<endl;
	    }
	  
// NB the full minimalization procedure ONLY works for locally soluble quartics
//    so we have to check solubility before minimalizing.
// But we can partially minimise anyway, which helps local solubility test
		  
	  m00=one;
	  m11=one; m12=zero; m21=zero; m22=one;
	      
	  if(xverb) cout << "preliminary minimalization of gg...\n";
	  minim_all(ga,gb,gc,gd,ge,ggI,ggJ,ggbadp,m00,m11,m12,m21,m22,0,xverb);
	  gg.assign(ga,gb,gc,gd,ge);  // must reset roots before searching
	  if(xverb) 
	    {
	      show_transform(m00,m11,m12,m21,m22);
	      cout<<"After preliminary minimalizing, gg = "<<gg<<endl;
	      cout<<"I(gg) = "<<ggI<<endl;
	      cout<<"J(gg) = "<<ggJ<<endl;
	    }
	  
	  if(locallysoluble(gg,plist,p))
	    {
	      if(xverb) 
		{
		  cout << "Everywhere locally soluble, ";
		  cout << "minimalizing gg...\n";
		}
	      minim_all(ga,gb,gc,gd,ge,ggI,ggJ,ggbadp,m00,m11,m12,m21,m22,1,xverb);
	      gg.assign(ga,gb,gc,gd,ge);  // must reset roots before searching
	      if(xverb) 
		{
		  show_transform(m00,m11,m12,m21,m22);
		  cout<<"After minimalizing, gg = "<<gg<<endl;
		  cout<<"I(gg) = "<<ggI<<endl;
		  cout<<"J(gg) = "<<ggJ<<endl;
		  cout<<"Now reducing gg...\n";
		}
	      if(verb>1) 
		cout<<"Descendent quartic (before reduction) is = "<<gg<<endl;
	      bigint oldga(abs(ga)); int better=1, ired;
	      bigint n00(m00), n11(m11), n12(m12), n21(m21), n22(m22);
	      for(ired=0; (ired<5)&&better; ired++)
		{
		  reduce(ga,gb,gc,gd,ge,n00,n11,n12,n21,n22);
		  better = (abs(ga)<oldga);
		  if(better)
		    {
		      oldga=abs(ga);
		      m00=n00; m11=n11; m12=n12; m21=n21; m22=n22;
		      gg.assign(ga,gb,gc,gd,ge);
		      if(verb>1) 
			cout<<"Descendent quartic (after "<<(ired+1)<<" reductions) is = "<<gg<<endl;
		    }
		}
	      if(xverb) 
		show_transform(m00,m11,m12,m21,m22);
	      if(verb) 
		cout<<"Descendent quartic (after reduction) is = "<<gg<<endl;
	      
 // Step 5: We have an ELS descendent quartic;
 //         now attempt to find a rational point on it

	      quartic_sieve qs(&gg,2,0);
	      if(verb) 
		cout << "Searching for points on gg up to height "<<hlim<<endl;
	      
//	      if(qs.search(hlim,1000000))
	      if(qs.search(hlim))
		{
		  bigint x2, y2, z2, x3, y3, z3;
		  qs.getpoint(x,y,z);
		  if(verb)
		    {
		      show_xyz(x,y,z);
		      cout<<"mapping this point back to original quartic:\n";
		    }
		  x3 = m11*x+m12*z;
		  z3 = m21*x+m22*z;
		  y3 = m00*y;

		  bigint q1xz = Q10*sqr(x3) + Q11*x3*z3 + Q12*sqr(z3);
		  bigint q3xz = Q30*sqr(x3) + Q31*x3*z3 + Q32*sqr(z3);

		  x2 = abs(q10*sqr(q1xz) + q11*q1xz*q3xz + q12*sqr(q3xz));
		  z2 = abs(q30*sqr(q1xz) + q31*q1xz*q3xz + q32*sqr(q3xz));
		  bigint fac=gcd(x2,z2);
		  if(fac>1) {x2/=fac; z2/=fac;}
		  y2 = d1*sqr(x2) + c*x2*z2 + d2*sqr(z2);
		  if(isqrt(x2,x)&&isqrt(z2,z)&&isqrt(y2,y))
		    {
		      if(verb) 
			{
			  cout<<"Point on original quartic is "; 
			  show_xyz(x,y,z);
			}
		    }
		  else
		    {
		      cout<<"Failed! x2="<<x2<<", z2="<<z2<<", y2="<<y2<<".\n";
		    }
		  return +1;
		}
	      else
		{
		  if(verb) cout << "No point found.\n";
		  return 0;
		}
	    }
	  else
	    {
	      if(verb) cout<<"Not locally soluble (p="<<p<<")\n";
	      return -1;
	    }
	}
      else
	{
	  if(verb) cout<<"q3=d3 not soluble.\n";
	  return -1;
	}
    }
  else // q1=d3 not soluble
    {
      if(verb) cout<<"q1=d3 not soluble.\n";
      return -1;
    }  
}
