//
// isogs.cc -- points from lattices, and isogeny routines
//

#include "isogs.h"

//#define DEBUG

#ifdef LiDIA
#define EPS1 0.0001
#else
#define EPS1 0.1 // slower than 0.001 but doesn't miss so many isogenies 
                 // when in ordinary double precision...
#endif

#define close(x,y,e) (abs((x)-(y))<(e))

CurveRedArray lisog(const CurveRed& CR, Cperiods& cp, long ell, int verbose)
//INPUT: a curve of type CurveRed, so we know its invariants & conductor
//       the periods of the curve in a Cperiods type
//       a prime ell (primality not checked)
//OUTPUT: an array of the curves isogenous to the given curve (possibly empty)
{
  bigcomplex x,X,y,Y,z;
  bigint a1,a2,a3,a4,a6, b2,b4,b6,b8;
  CR.getai(a1,a2,a3,a4,a6);
  CR.getbi(b2,b4,b6,b8);
  bigfloat ra1=I2bigfloat(a1),ra2=I2bigfloat(a2),ra3=I2bigfloat(a3),
         ra4=I2bigfloat(a4),ra6=I2bigfloat(a6),
         rb2=I2bigfloat(b2),rb4=I2bigfloat(b4),rb6=I2bigfloat(b6); 
  // b8 isn't used
  if (verbose>1)
    cout<<"\nra1: "<<ra1<<"\tra2: "<<ra2<<"\tra3: "<<ra3<<"\tra4: "<<ra4<<"\tra6: "<<ra6<<"\nrb2: "<<rb2<<"\trb4: "<<rb4<<"\trb6: "<<rb6<<endl;
  bigint conductor = getconductor(CR);
  bigcomplex w1, w2;  cp.getwi(w1, w2);
  int type = getnorm_code(cp);  //Must do this BEFORE doing norm_region
                                //as w1, w2 should be lattice-normed
//N.B. This function needs w1,w2 to be lattice-normalised, i.e. either
//type=1 and Re(tau)=0.5 or type=2 and Re(tau)=0.  But the XY_coords function
//needs it to be region-normalised, i.e. tau in fundamental region.
  if ( (type==0) || (type==3) )
    {cerr<<" ## lisog: called with Cperiods not a lattice type, return null"
         <<endl;
     return 0; } // will construct array of length 0 from this
  Cperiods cpregion(cp);
  cpregion.norm_region(); // ie for tau=w2/w1 in fundamental region
                          // needed for XYcoords
  long subgroup, nsubgroups;
  if (ell==2)
    { if (type==1) nsubgroups = 1; else nsubgroups = 3; }
  else nsubgroups = 2;
  // Now nsubgroups is the number of subgroups of C/L of index ell 
  // defined over R
  CurveRedArray ans(nsubgroups);  // = max number of l-isog curves
  long ncurves=0;

  // other loop variables:
  bigcomplex t, w, ti, ui;
  bigint ad4, ad6, lad4, lad6;
  bigfloat rad4, rad6;
  bigint ell2 = ell*ell;
  bigint ell3 = ell*ell2;
  bigint ell4 = ell2*ell2;
  bigint ell6 = ell2*ell4;

  for (subgroup = 1; subgroup <= nsubgroups; subgroup++)
  {
    if (ell==2)
      { if (subgroup==1) z = w1/2;
	else if (subgroup==2) z = w2/2;
	else z = (w1 + w2)/2; }
    else
      { if (subgroup==1) z = w1/ell;
	else if (type==1) z = (w1 - w2 - w2)/ell;
	else z = w2/ell; }
    t =  w = 0;
    if (verbose>1) cout<<"Subgroup: "<<subgroup<<": z = "<<z<<endl;

    long iz, ilim;
    if (ell==2) ilim = 1;
    else ilim = (ell - 1)/2;
    for (iz=1; iz <= ilim; iz++)
      { cpregion.XY_coords(X, Y, iz*z);
	if(ell==2) Y=0;  // fix to avoid q(z)=-q(tau) awkward case
//Must convert from Y^2=4X^3 + ... model to our minimal model:	
	x = X - (ra1*ra1 + 4*ra2)/12;
	y = (Y - ra1*x - ra3)/2;
	if (verbose>2)
	  cout<<"i = "<<iz<<":  i*z = " << iz*z << "\n x = "<<x<<"\n  y = "<<y<<endl ;
	if (ell==2) ti = 3*x*x + 2*ra2*x + ra4 - ra1*y;
	else ti = 6*x*x + rb2*x + rb4;
	ui = 4*x*x*x + rb2*x*x + 2*rb4*x + rb6;
	t += ti;
	w += ui + x*ti;
      }
    if (verbose>1)
      cout<<"t = "<<t<<";\tw = "<<w<<endl;

    rad4 = ra4 - 5*real(t);
    rad6 = ra6 - rb2*real(t) - 7*real(w);
    if (verbose>1)
      cout<<"new a4: "<<rad4<<"\tnew a6: "<<rad6<<endl;

    ad4 = Iround(rad4);
    ad6 = Iround(rad6);
    if (verbose>1)
      cout<<"bigint values are "<<ad4<<" and "<<ad6<<endl;

    if ( close(I2bigfloat(ad4) , rad4 , EPS1) &&  
	 close(I2bigfloat(ad6) , rad6 , EPS1) )
      {
	Curve newcurve(a1,a2,a3,ad4,ad6);
	if(verbose>0) cout << "Testing curve " << newcurve << endl;
	Curvedata newCD(newcurve,1);
	if ((Curve)CR == (Curve)newCD)
	  cout<<" ## Warning! curve possibly "<<ell<<"-isogenous to itself"<<endl;
	else
	  {
	    CurveRed newCR(newCD);
	    if (conductor==getconductor(newCR))
	      { 
		ans[ncurves++]=newCR;
		if (verbose>1) cout<<"new curve:\n"<<(Curve)newCD<<endl;
	      }
	  }
      }
    else
      {
	if(verbose>1) cout << "Not close enough\n";
	rad4 = rad4*I2bigfloat(ell4);
	rad6 = rad6*I2bigfloat(ell6);
	if (verbose>1)
	  cout<<"new a4: "<<rad4<<"\tnew a6: "<<rad6<<endl;
	lad4 = Iround(rad4);
	lad6 = Iround(rad6);
	if (verbose>1)
	  cout<<"bigint values are "<<lad4<<" and "<<lad6<<endl;
	if ( close(I2bigfloat(lad4) , rad4 , EPS1) &&
	     close(I2bigfloat(lad6) , rad6 , EPS1) )
	  { 
	    Curve newcurve(ell*a1,ell2*a2,ell3*a3,lad4,lad6);
	    if(verbose>0) cout << "Testing curve " << newcurve << endl;
	    Curvedata newCD(newcurve,1);
	    if ((Curve)CR == (Curve)newCD)
	      cout<<" ## Warning! curve possibly "<<ell<<"-isogenous to itself"<<endl;
	    else 
	      {
		CurveRed newCR(newCD);
		if (conductor==getconductor(newCR))
		  { 
		    ans[ncurves++]=newCR;
		    if (verbose>1) cout<<"new curve:\n"<<(Curve)newCD<<endl;
		  }
	      }
	  }
	else if(verbose>1) cout << "Not close enough\n";
      }
  } // ends for-loop

  ans.truncate(ncurves);
  return ans;
}

int semistable(const CurveRed& CR)
{
  int ans=1;
  bigintArray plist = getbad_primes(CR);
  for(bigintvar pvar(plist); ans&&pvar.ok(); pvar++)
    ans = ans && getord_p_N(CR,pvar.value())<2;
  return ans;
}

longlist getelllist(const CurveRed& CR)
{
static const bigrational j11a(-32768);
static const bigrational j11b(-121);
static const bigrational j11c(-24729001);

static const bigrational j17a(-297756989,2);
static const bigrational j17b(-882216989,131072);


static const bigrational j37a(-9317);
static const bigrational j37b(atoI("-162677523113838677"));

static const bigrational j19(-884736);
static const bigrational j43(-884736000);
static const bigrational j67(atoI("-147197952000"));
static const bigrational j163(atoI("-262537412640768000"));

  longlist ans(6);  //upper bound -- truncated before return
  int num=0;ans[num++]=2;ans[num++]=3;ans[num++]=5;ans[num++]=7;
  bigint N = getconductor(CR);
  if(!semistable(CR))
    {
      ans[num++]=13;
      bigint c4=getc4(CR), delta=getdiscr(CR);
      bigint c43=pow(c4,3);
      bigrational jay(c43,delta);
      if((jay==j11a)||(jay==j11b)||(jay==j11c)) ans[num++]=11; 
      else {
	  if((jay==j37a)||(jay==j37b)) ans[num++]=37; 
	  else {
	    if(jay==j19) ans[num++]=19; 
	    else {
	      if(jay==j43) ans[num++]=43; 
	      else {
		if(jay==j67) ans[num++]=67; 
		else {
		  if(jay==j163) ans[num++]=163; 
		  else {
		    if(div(17,N)&&((jay==j17a)||(jay==j17b))) ans[num++]=17;
		  }
		}
	      }
	    }
	  }
	}
    }
  ans.truncate(num);
  return ans;
}

IsogenyClass::IsogenyClass(const CurveRed& C, int verbose)
{
  verb=verbose;
  cp = Cperiods(C);
  if(verb)
    {
      cout << endl;
      C.display(cout);
      cout<<"\nPeriod lattice:\n" << cp << endl;
    }
  llist = getelllist(C);
  ss = semistable(C);
  if(verb)
    {
      cout << "Curve is ";if(!ss)cout<<"NOT ";cout<<"semistable."<<endl;
    }
  nell = llist.length;
  curves   = CurveRedArray(MAXNCURVES);
  fromlist = longlist(MAXNCURVES);
  isoglist = longlist(MAXNCURVES);
  curves[0]=C; fromlist[0]=0; isoglist[0]=0;
  matij = longlist(MAXNCURVES*MAXNCURVES);
 // initialize to 0:
  for(long ij=0; ij<MAXNCURVES*MAXNCURVES; ij++) matij[ij]=0;
}

void IsogenyClass::process(long i)  // process i'th curve
{
  longlist lworks(nell);
  CurveRed thisc = curves[i];
  if (verb) cout << "Working on curve " << i+1 << ": " << (Curve)thisc << endl;
  Cperiods pers(thisc);
  for (long il=0; il<nell; il++)
    {
      long ell = llist[il];
      if (verb) cout << "trying l = " << ell << "..." << flush;
      CurveRedArray lisogcurves = lisog(thisc,pers,ell,verb);
      long number =  lisogcurves.getlength();
      if (verb) cout << number << " isogenous curves found." << endl;
      if(i==0) lworks[il] = (number>0);
      for (long n=0; n<number; n++)
	{ 
	  CurveRed newc = lisogcurves[n];
	  if (verb) cout << "\t"<<n+1<<": "<<(Curve)newc<<"\t: ";
	  int isnew = 1;
	  for (long j=0; (j<ncurves); j++)
	    {
	      if ((Curve)newc==(Curve)curves[j])
		{
		  isnew=0;
		  matset(i,j,ell);
		  matset(j,i,ell);
		}
	    }
	  if (isnew) 
	    {
	      curves[ncurves] = newc;
	      fromlist[ncurves]=i;
	      isoglist[ncurves]=ell;
	      matset(i,ncurves,ell);
	      matset(ncurves,i,ell);
	      ncurves++;
	      if (verb) cout << "new # " << ncurves << endl;
	    }
	  else if (verb) cout << "repeat" << endl;
	}
    }  // end of ell loop
  if(i==0) // reset llist to good l only;
    {
      long ngood=0;
      for(long i=0; i<nell; i++) 
	if(lworks[i]) 
	  llist[ngood++]=llist[i];
      nell=ngood;
      llist.truncate(nell);
      if(verb)
	{
	  cout << "Number of useful l is " << nell;
	  if(nell) cout << ": " << llist;
	  cout<<endl;
	}
    }
}

void IsogenyClass::grow(void)       // does the work
{
  if(verb) cout << "Trying l values: " << llist << endl;
// N.B. ncurves will increase as more curves are found
  for (ndone=0, ncurves=1; ndone<ncurves; ndone++) 
    {
//      cout << "After processing "<<ndone<<" curve(s), :";
//      display(cout);
      process(ndone);
    }
  curves.truncate(ncurves);
}

void IsogenyClass::displaycurves(ostream& os)const
{
  os << endl << ncurves << " curve(s) in the isogeny class"<<endl<<endl;
  if(ncurves==0) return;
  long i,j;
  for (i=0; i<ncurves; i++)
    { Curve ci = (Curve)curves(i);
      os << (i+1) << ": " << ci;
      if (i>0) os << "  is "<< isoglist(i)<<"-isogenous to curve "<<fromlist(i)+1;
      os<<endl;
    }
  os<<endl;
}

void IsogenyClass::displaymat(ostream& os)const
{
  if(ncurves==0) return;
  long i,j;
  os << "Isogeny matrix:\n";
  os << "\t"; for(j=0; j<ncurves; j++) os<<(j+1)<<"\t";  os<<"\n";
  for(i=0; i<ncurves; i++)
    {
      os<<(i+1)<<"\t"; for(j=0; j<ncurves; j++) os<<mat(i,j)<<"\t";  os<<"\n";
    }
  os<<endl;
}

void IsogenyClass::dumpdata(ostream& os, long rank)  
// output for textab to input
{
  os << ncurves << "\n";
  long ic, jc, ip, np, il, nj;
  char sep = ' ';
  for (ic=0; ic<ncurves; ic++)
    {
      CurveRed& C = curves[ic];
      os << C.a1 << sep << C.a2 << sep << C.a3 << sep << C.a4 << sep 
         << C.a6 << sep ;
      os << rank << sep;
//cout << "C.ntorsion = " << C.ntorsion << endl;
      os << C.get_ntorsion() << sep;
//cout << "After C.get_ntorsion(), C.ntorsion = " << C.ntorsion << endl;
      np = C.bad_primes.length;
      if(sign(C.discr)>0) os << "+1" <<sep;
      else  os << "-1" <<sep;
      for(ip=0; ip<np; ip++) os << C.reduct_array[ip].ord_p_discr << sep;
      for(ip=0; ip<np; ip++) os << C.reduct_array[ip].ord_p_j_denom << sep;
      for(ip=0; ip<np; ip++) os << C.reduct_array[ip].c_p << sep;
      for(ip=0; ip<np; ip++) os << C.reduct_array[ip].Kcode.code << sep;
      os << nell << sep;
      for(il=0; il<nell; il++)
	{
	  long ell = llist[il];
	  for(jc=0, nj=0; jc<ncurves; jc++) 
	    if(mat(ic,jc)==ell) nj++;
	  os << nj << sep;   // # of ell-isogenous curves
	  os << llist[il] << sep;
	  for(jc=0; jc<ncurves; jc++) 
	    if(mat(ic,jc)==ell) os << (jc+1) << sep;
                    	    // id #s of l-isogenous curves
	}
      os << endl;
    }
}

longlist IsogenyClass::getmat() const
{
  longlist ans(ncurves*ncurves);
  long i,j;
  for(i=0; i<ncurves; i++)
    for(j=0; j<ncurves; j++)
      ans[i*ncurves+j] = mat(i,j);
  return ans;
}


