#include <builtin.h>
#include "compproc.h"
#include "points.h"
#include "log2.h"
#include "mquartic.h"
#include "msoluble.h"
#include "mrank2.h"
#include "sqfdiv.h"
#include "desc2.h"

//#define SELMER_ONLY  // obsolete testing option, not supported 
                       // would need fixing to give correct output
                       // but the first rank upper bound output is correct
#define USE_SECOND_DESCENT

//#define DEBUG_ROOTS
//#define DEBUG_SQF
//#define DEBUG_ELS
//#define DEBUG_GLS
#define MAX_R 5   // will not attempt to list all coset reps
                  // for 2E(Q) in E(Q) if rank is more than this

void rank2::makepoint(const bigint& c,const bigint& d1,const bigint& d2,
	       const bigint& x, const bigint& y, const bigint& z, 
	       int which)
{
  Point P(ee);
  if (verbose) cout<<" (x:y:z) = ("<<x<<":"<<y<<":"<<z<<")\n";
  if(which==0)
    {
      if(sign(z)!=0)	P.init(ee, d1*x*x*z, d1*x*y, pow(z,3));
      if(verbose) 
	{
	  cout<<"\tCurve E " /* <<(Curve)ee */ <<"\tPoint "<< P;
	  bigfloat h = height(P);
	  cout << ", height = " << h;
	  if(!P.isvalid()) {cout << " --warning: NOT on curve! " ;}
	  cout << "\n";
	}
    }
  else
    {
      if(verbose) // display point on isogenous curve too
	{
	  Point Q(eedash);
	  if(sign(z)!=0) Q.init(eedash, d1*x*x*z, d1*x*y, pow(z,3));
	  cout<<"\tCurve E' " /* <<(Curve)eedash */ <<"\tPoint "<< Q;
	  bigfloat h = height(Q);
	  cout << ", height = " << h;
	  if(!Q.isvalid()) {cout << " --warning: NOT on curve! " ;}
	  cout << "\n";
	}
      const bigint& xz=x*z, x2=x*x, z2=z*z;
      if(sign(xz)!=0) P.init(ee,2*y*y*xz,y*(d1*x2*x2-d2*z2*z2),pow(2*xz,3));
      if(verbose) 
	{
	  cout<<"\tCurve E " /* <<(Curve)ee */ <<"\tPoint "<< P;
	  bigfloat h = height(P);
	  cout << ", height = " << h;
	  if(!P.isvalid()) {cout << " --warning: NOT on curve! " ;}
	  cout << "\n";
	  
	}
    }
  if(order(P)<0)  pointlist[npoints++] = P; // else torsion so ignore
}

int rank2::testquartic(const bigint& c,const bigint& d1,const bigint& d2,int which)
{
  // creates quartic (d1,0,c,0,d2), assumed els, and tries to find a rational point
  // returns +1 if rational point found (handled my makepoint())
  //          0 if undecided
  quartic q(d1,  0, c,  0, d2);
  if (verbose) cout<<q<<": ";

  bigint x,y,z;

  // First a quick search for a small point:
  if (ratpoint(q,1,lim1,x,y,z))
    { 
      makepoint(c,d1,d2,x,y,z,which);
      return 1;
    }

  // Next a more serious search using a sieve,
  // but not too far if we are going to use second descent
  quartic_sieve qs(&q,2,0);
  long lim3=lim2; 
#ifdef USE_SECOND_DESCENT
  if(lim3>6) lim3=6;      // save the long search till later...
#endif
  if(qs.search(lim3, 1, 1)) // maxnpoints, pos-x-only
    { 
      qs.getpoint(x,y,z);
      makepoint(c,d1,d2,x,y,z,which);
      return 1;     
    }
  if (verbose) 
    cout<<" no rational point found (hlim="<<lim3<<")\n";
  return 0;
}  /* end of testquartic*/
 
int rank2::second_descent(const bigint& c, const bigint& d1, const bigint& d2, int which)
  // creates quartic (d1,0,c,0,d2), assumed els, and tries to find a rational point
  // by second descent (after testquartic() has failed using direct search)
  // returns +1 if rational point found (handled by makepoint())
  //          0 if undecided
  //         -1 if no rational point exists (proved by second descent)
{
  int res, verb=verbose; if(verb) verb--;  // reduced verbosity level within desc2()
  bigint x,y,z;
  if(verbose) cout<<"d1="<<d1<<": "<<flush;
  if(which)
    res = desc2(c,d1,d2,badprimes, supp0, elsgens0, mask0,  lim2,x,y,z,verb);
  else
    res = desc2(c,d1,d2,badprimes, supp1, elsgens1, mask1,  lim2,x,y,z,verb);

  if(verbose==1) cout<<endl;

  switch(res)
    {
    case 1: // Positive success, point found
      makepoint(c,d1,d2,x,y,z,which);
      if(verbose)
	cout<<"Second descent successfully found rational point for d1="<<d1<<"\n\n";
      break;
    case -1: // Negative success, no point exists
      if(verbose)
	cout<<"Second descent shows that no rational point exists for d1="<<d1<<"\n\n";
      break;
    case 0: // Undecided
      if (verbose) 
	cout<<"Second descent inconclusive for d1="<<d1<<": ELS descendents exist but no rational point found\n\n";
    }
  return res;
}

void rank2::find_elsgens(int which, const bigint& c, const bigint& d)
{
  if (verbose>1) 
    {
      if(which) cout<<"\n";
      cout<<"Finding els gens for E"; if(which) cout<<"'";
      cout<<" (c"; if(which) cout<<"'";
      cout<<"= "<<c<<", d"; if(which) cout<<"'";
      cout<<"= "<<d<<")"<<endl;
    }
  bigint ddash = c*c-4*d;
  int posd1, istep=1;
  if (sign(ddash)< 0) 
    posd1=1; 
  else 
    posd1= (c+sqrt(ddash)< 0);
  if(posd1) istep=2;  // this will skip negative d1
  
  int extra2torsion=0, i2t, n2t=1;
  bigint ee2, ee3;
  if(which)
    {
      if(d_is_sq)
	{
	  extra2torsion=1; n2t=3;
	  ee2 = e2dash;
	  ee3 = e3dash;
	}
    }
  else
    {
      if(ddash_is_sq)
	{
	  extra2torsion=1; n2t=3;
	  ee2 = e2;
	  ee3 = e3;
	}      
    }

  bigint d1, d2, badp;
  bigintArray supp = support(d);

  long ns = supp.length;
  if(verbose>1)
    {
      cout<<"Support (length "<<ns<<"): "<<supp<<endl;
    }
  if(ns>31) {cerr<<"Too many primes dividing d!\n"; exit(1);}

  long mask=0, maxn=1<<ns, index, j, nelsgens=0;
  bigintArray elsgens(ns);

// First process 2-torsion
//  n2t=0; /////////////////////to ignore 2-torsion for testing/////////////
  bigint d1x;
  for(i2t=0; i2t<n2t; i2t++)
    {
      switch(i2t) {
      case 0: d1=d; break;
      case 1: d1=ee2; break;
      case 2: d1=ee3; break;
      }
#ifdef DEBUG_ELS
      if(verbose>1) cout<<"Processing 2-torsion d1 = " << d1 << endl;
#endif
      index=makeindex(supp, d1, d1x);
      d1=d1x;  // the square-free part, = d1/square
      if(mask&index) continue;
      if(index)
	{
	  elsgens[nelsgens++]=d1;
	  if(verbose>1) 
	    cout<<"Adding (torsion) els generator #"<<nelsgens<<": d1 = " << d1;
	  for(j=ns-1; j>=0; j--)
	    if(testbit(index,j)) {setbit(mask,j); break;}
#ifdef DEBUG_ELS
	  if(verbose>1) cout<<" (pivot = "<<j<<": "<<supp(j)<<")";
#endif
	  if(verbose>1) cout<<endl;
	}
    }
// Record that the first 0, 1 or 2 elsgens come from known torsion points, 
// so can be ingored in the gls step later:
  if(which)
    {
      nt2gens1=nelsgens;
    }
  else
    {
      nt2gens0=nelsgens;
    }

// Now systematically go through square-free divisors d1 of d, always modulo 
// the subgroup of those which we know are els

  for(index=istep; index<maxn; index+=istep)
    {
      if(mask&index) continue;
      d1 = makenum(supp,index);
      d2 = d/d1;
#ifdef DEBUG_ELS
      if(verbose>1) cout<<"Testing d1 = "<<d1<<"\t";
#endif
      int res = locallysoluble(d1,c,d2,badprimes,badp);
      if(res)
	{
	  elsgens[nelsgens++]=d1;
	  if(verbose>1) cout<<"Adding els generator #"<<nelsgens<<": d1 = " << d1;
	  for(j=ns-1; j>=0; j--)
	    if(testbit(index,j)) {setbit(mask,j); break;}
#ifdef DEBUG_ELS
	  if(verbose>1) cout<<" (pivot = "<<j<<": "<<supp(j)<<")";
#endif
	  if(verbose>1) cout<<endl;
	}
      else 
	{
#ifdef DEBUG_ELS
	  if(verbose>1) cout<<"not locally soluble at p = "<<badp<<endl;
#endif
	}
    }
  elsgens.truncate(nelsgens);
  if(verbose>1)
    {
      cout<<"After els sieving, nelsgens = " << nelsgens;
      cout << endl;

      cout<<"2-rank of S^{phi";      if(which) cout<<"'";
      cout<<"}(E";                   if(!which) cout<<"'";
      cout<<") = "<<nelsgens<<endl;

      if(nelsgens>0) cout<<"gens: "<<elsgens<<endl;
      //      cout<<"mask = "<<mask<<endl;
    }
  if(which)
    {
      supp1=supp; elsgens1=elsgens; mask1=mask; els1=nelsgens;
    }
  else
    {
      supp0=supp; elsgens0=elsgens; mask0=mask; els0=nelsgens;
    }
}

void rank2::find_glsgens(int which, const bigint& c, const bigint& d)
{
  bigintArray& elsgens = (which? elsgens1: elsgens0);
  long nelsgens        = (which? els1: els0);
  long nt2gens         = (which? nt2gens1: nt2gens0);
  bigintArray glsgens(nelsgens);
  long* piv_pos = new long[nelsgens];
  long* piv_gen = new long[nelsgens];
  long j, pivj, maxn = 1<<nelsgens;
  longlist nonglslist(maxn); long nnongls=0;
  long nglsgens=0, nglsgens2=0, sha_contrib=0;
  long mask, pivmask, index, stage, nstages=1;
  long shortfall1, shortfall2, es2, ns2;
  bigint d1, d2;
  int res;
  
#ifdef USE_SECOND_DESCENT
  nstages=2;
#endif

// first record the 2-torsion contribution:

  for(index=0; index<nt2gens; index++)
    {
      mask=(1<<index);
      d1=elsgens(index);
      for(j=0; j<nglsgens; j++)
	{
	  if(testbit(mask,piv_pos[j])) mask=mask^piv_gen[j];
	}
      if(!mask) 
	{
#ifdef DEBUG_GLS
	  cout<<"d1="<<d1<<": known gls (2-torsion)"<<endl;
#endif	      
	  continue; // as we are certainly in the gls subgroup
	}
      // Mask has a bit set, find it for use as next pivot
      for(j=nelsgens-1; j>=0; j--) if(testbit(mask,j)) {break;}
      pivj=j; pivmask=mask;
      glsgens[nglsgens]=d1;
      if(verbose>1) 
	{
	  cout<<"Adding (2-torsion) gls generator #"<<(nglsgens+1)<<": d1 = " << d1;
	}
      piv_pos[nglsgens]=pivj;
      piv_gen[nglsgens]=pivmask;
#ifdef DEBUG_GLS
      if(verbose>1) 
	cout<<" (piv_pos = "<<pivj<<", piv_gen = "<<index<<")";
#endif
      if(verbose>1) cout<<endl<<endl;
      nglsgens++;
    }

// Next test all first descent curves for global solubility (stage 1)
// and (optionally) do a second descent (stage 2) on unsuccessful ones.

// The next lines are in case 2-torsion accounts for everything, 
// as then the following loop is not executed at all: 
  shortfall1 = 0;
  if(which)
    {gls1=gls21=nglsgens;}
  else
    {gls0=gls20=nglsgens;}

  for(stage=1; (stage<=nstages)&&(nglsgens<nelsgens); stage++)
    {
#ifdef USE_SECOND_DESCENT
      if(verbose) 
	{
	  if(stage==1) 
	    cout << "First stage (no second descent yet)...\n";
	  else        
	    cout << "Second stage (using second descent)...\n";
	}
#endif
      for(index=1; (index<maxn)&&(nglsgens<nelsgens); index++)
	{
#ifdef DEBUG_GLS
	  d1 = makenum(elsgens,index);  // do it now for output purposes
#endif
	  mask=index;
	  for(j=0; j<nglsgens; j++)
	    {
	      if(testbit(mask,piv_pos[j])) mask=mask^piv_gen[j];
	    }
	  if(!mask) 
	    {
#ifdef DEBUG_GLS
	      cout<<"d1="<<d1<<": known gls"<<endl;
#endif	      
	      continue; // as we are certainly in the gls subgroup
	    }

// Mask has a bit set, find it for possible use as next pivot
	  for(j=nelsgens-1; j>=0; j--) if(testbit(mask,j)) {break;}
	  pivj=j; pivmask=mask;

// Test if this value is equivalent modulo current gls subgroup
// to one known to be NOT gls by a previous 2nd descent
	  for(j=0; (j<nnongls)&&mask; j++)
	    {
	      if(mask==nonglslist(j)) mask=0;
	    }
	  if(!mask) 
	    {
#ifdef DEBUG_GLS
	      cout<<"d1="<<d1<<": known NOT gls"<<endl;
#endif	      
	      continue; // we are certainly NOT in the gls subgroup
	    }

#ifndef DEBUG_GLS
	  d1 = makenum(elsgens,index);  // else we did it earlier
#endif
	  d2 = d/d1;
	  if(stage==1) res = testquartic(c,d1,d2,which);
	  else         res = second_descent(c,d1,d2,which);
#ifdef DEBUG_GLS
	  cout<<"d1="<<d1<<": res = "<<res<<endl;
#endif	      
	  switch(res){
	  case 0:
	    {
	      if(stage==2) 
		sha_contrib++;  // We have a potential element of Sha
	      break;
	    }
	  case 1:
	    {
	      glsgens[nglsgens]=d1;
	      if(verbose>1) cout<<"Adding gls generator #"<<(nglsgens+1)<<": d1 = " << d1;
	      piv_pos[nglsgens]=pivj;
	      piv_gen[nglsgens]=pivmask;
#ifdef DEBUG_GLS
	      if(verbose>1) cout<<" (piv_pos = "<<pivj<<", piv_gen = "<<index<<")";
#endif
	      if(verbose>1) cout<<endl<<endl;
	      nglsgens++;
	      break;
	    }
	  case -1:
	    {
	      nonglslist[nnongls++] = pivmask;
	    }
	  } // end of switch(res)
	}   // end of index loop

      if(stage==1) // finished stage 1, i.e. first descent
	{
	  shortfall1 = nelsgens-nglsgens;
	  if(verbose)
	    {
	      if(nstages==2) cout << "After first descent, ";
	      cout << "shortfall in rank = " << shortfall1 << "\n";
	      if(verbose>1) cout<<"\n";
	    }
	  if(which)
	      gls1=gls21=nglsgens;
	  else
	      gls0=gls20=nglsgens;
	}
      else // finished stage 2, i.e. 2nd descent
	{
	  if(which)
	    gls21=nglsgens;
	  else
	    gls20=nglsgens;
	}

    }  // end of stage loop
  
  glsgens.truncate(nglsgens);
  if(which)
      glsgens1=glsgens; 
  else
      glsgens0=glsgens; 

  es2 = nglsgens;
#ifdef USE_SECOND_DESCENT
  ns2 = (1<<es2);
  if(sha_contrib>0)
    {
      ns2 += sha_contrib;
      long kns2=ns2;
      if (!intlog2(ns2,es2,1))
	{
	  cout<<"\n!!! ns2 = "<<kns2<<" not a power of 2, rounding up to "<<ns2<<"\n";
	}
    }
  shortfall2 = es2-nglsgens;
  if(which)
    els21=es2;
  else
    els20=es2;
#else
  if(which)
    els21=els1;
  else
    els20=els0;
#endif
  
  if(verbose>1)
    {
      cout<<"After gls sieving, nglsgens = " << nglsgens << endl;
      if(nelsgens>es2)
	cout<<"rank upper bound reduced by " << (nelsgens-es2) 
	  << " via second descent\n";
      cout<< "shortfall in rank from first  descent = "<<shortfall1<<endl;
#ifdef USE_SECOND_DESCENT
      cout<< "shortfall in rank from second descent = "<<shortfall2<<endl;
#endif

      if(nglsgens>0) cout<<"gens: "<<glsgens<<endl;
    }

  delete[] piv_pos;
  delete[] piv_gen;
}
 
rank2::rank2(Curvedata* ec, int verb, long l1, long l2)
{
  the_curve = ec;
  bigint a1, a2, a3, a4, a6;
  ec->getai(a1,a2,a3,a4,a6);
  pointlist =  PointArray(256);
  fullnpoints = npoints = 0;
  lim1 = l1;  lim2 = l2; verbose=verb;
  rank = 0;            // default value if failure occurs
  bigint s2,s4,s6;
  success = 1;
  int scaled=0;
  if (odd(a1) || odd(a3)) 
    { s2= a1*a1+ 4*a2;
      s4=  8*(a1*a3+ 2*a4);
      s6= 16*(a3*a3+ 4*a6);
      scaled=1;
    }
  else 
    { s2=a2; s4=a4; s6=a6;
    }

  bigintArray xlist = Introotscubic(s2,s4,s6);
  ntwo_torsion = xlist.length;
  if (ntwo_torsion==0)
    { 
      success=0;
      if (verbose) cout << "No points of order 2\n";
      return;
    }

  if(verbose) cout << ntwo_torsion << " points of order 2:\n";
  two_torsion=PointArray(ntwo_torsion);
  for(int n=0; n<ntwo_torsion; n++)
    {
      bigint ei = xlist[n];
      if(scaled) two_torsion[n].init(the_curve,2*ei,-a1*ei-4*a3,8);
      else two_torsion[n].init(the_curve,ei,0,1);
      if(verbose)
	{ if(n>0) cout<<", "; cout<<two_torsion[n];}
    }
  if(verbose)cout<<endl;
  const bigint& x0 = xlist[0];
  const bigint& c =  3 * x0 + s2;
  const bigint& d = x0*(c+s2) + s4;
  const bigint& cdash = - 2 * c;
  const bigint& ddash = c*c -  4*d;
  const bigint& disc = 2*d*ddash;
  if (is_zero(disc)) 
    {
      cout << "Curve is singular\n"; 
      success = 0;
      return;
    }

  ee     = Curvedata(0,  c,  0,  d,  0);
  eedash = Curvedata(0,cdash,0,ddash,0);
  if(verbose) 
    cout<<"Using 2-isogenous curve "<<(Curve)(Curvedata(eedash,1))
	<<endl;
  badprimes = pdivs(disc);
  
  bigint rootd; 
  if(isqrt(ddash,rootd))
  {
    ddash_is_sq=1;
    e2 = (-c-rootd)/2;
    e3 = (-c+rootd)/2;
  }
  if(isqrt(d,rootd))
  {
    d_is_sq=1;
    e2dash = c-2*rootd;
    e3dash = c+2*rootd;
  }

  long mw0,mw1,sel0,sel1,sha0,sha1,sel20,sel21,sha20,sha21,lindex;
  
  if(verbose) 
    {
      cout<<"-------------------------------------------------------\n";
      cout<<"First step, determining Selmer group\n";
      cout<<"-------------------------------------------------------\n";
    }
  find_elsgens(0,c,d);
  find_elsgens(1,cdash,ddash);

  rank_bound = els0+els1-2;
  
  if(verbose) 
    {
      cout<<"-------------------------------------------------------\n";
      cout<<"Rank <= "<<rank_bound<<"\n";
      cout<<"-------------------------------------------------------\n";
    }

#ifndef SELMER_ONLY
  if(verbose) 
    {
      cout<<"Second step, determining E(Q)/phi(E'(Q)) and E'(Q)/phi'(E(Q))\n";
      cout<<"-------------------------------------------------------\n";
      cout<<"1. E(Q)/phi(E'(Q))\n";
      cout<<"-------------------------------------------------------\n";
    }
  
  find_glsgens(0,c,d);
  
  npoints1=npoints;
  if(verbose) 
    {
      cout<<"-------------------------------------------------------\n";
      cout<<"2. E'(Q)/phi'(E(Q))\n";
      cout<<"-------------------------------------------------------\n";
    }
  find_glsgens(1,cdash,ddash);
#endif

  pointlist.truncate(npoints);

//Debug only:
//cout<<"gls20 = "<<gls20<<endl;
//cout<<"gls21 = "<<gls21<<endl;
//cout<<"els0 = "<<els0<<endl;
//cout<<"els1 = "<<els1<<endl;
//cout<<"els20 = "<<els20<<endl;
//cout<<"els21 = "<<els21<<endl;

  rank = gls20+gls21-2;         // lower bound for rank
  rank_bound  = els20+els21-2; //  = upper bound for rank

  mw0 = 1<<gls20;  
  sel0 = 1<<els0;    sha0=sel0/mw0;      
  sel20 = 1<<els20;  sha20=sel20/mw0;

// gls20  is the 2-rank of E/phi(E')                    (lower bound)
// els0   is the 2-rank of S^(phi)(E')                  (exact)
// els20  is the 2-rank of its subgroup phi'(S^(2)(E))  (exact if 2nd descent)
// els0-gls20  is the 2-rank of III(E')[phi)                 (upper bound)
// els20-gls20 is the 2-rank of its subgroup phi'(III(E)[2]) (upper bound)

  mw1 = 1<<gls21;  
  sel1 = 1<<els1;  sha1=sel1/mw1;
  sel21 = 1<<els21;  sha21=sel21/mw1;

// gls21   is the 2-rank of E'/phi'(E)                   (lower bound)
// els1    is the 2-rank of S^(phi')(E)                  (exact)
// els21   is the 2-rank of its subgroup phi(S^(2)(E'))  (exact if 2nd descent)
// els1-gls21  is the 2-rank of III(E)[phi']                 (upper bound)
// els21-gls21 is the 2-rank of its subgroup phi(III(E')[2]) (upper bound)

  index2=mw0*mw1;
  lindex=gls20+gls21;
  if(!ddash_is_sq) {index2/=2; lindex-=1;}
  
  int certain0 = (els20==gls20); // i.e. descent was conclusive
  int certain1 = (els21==gls21); //
  certain = certain0&&certain1;
  
  if (verbose||!certain) 
    {
      cout<<"\n";
      cout<<"-------------------------------------------------------\n";
      cout<<"Summary of results:\n";
      cout<<"-------------------------------------------------------\n";
      
      if(certain) cout << "\trank(E) " << "= " << rank;
      else        cout << "\t"<<rank<< " <= rank(E) <= " << rank_bound;
      cout<<"\n";
      cout << "\t#E(Q)/2E(Q) ";
      if(!certain) cout<<">";
      cout<<"= " << index2 << "\n\n";
      
      long lb, ub; int lb1, eq;
#ifdef USE_SECOND_DESCENT
      cout<<"Information on III(E/Q):\n";

      lb = 1; lb1=1; ub = sha1; eq=(lb==ub);
      cout<<"\t#III(E/Q)[phi']    ";
      if(eq) cout<<"= "<<ub<<"\n";
      else
	{
	  if(lb1) cout<<"<= "<<ub<<"\n";
	  else    cout<<"is between "<<lb<<" and "<<ub<<"\n";
	}

      lb = sel1/sel21; lb1=(lb==1); ub = sha1*sha20; eq=(lb==ub);
      cout<<"\t#III(E/Q)[2]       ";
      if(eq) cout<<"= "<<ub<<"\n";
      else
	{
	  if(lb1) cout<<"<= "<<ub<<"\n";
	  else    cout<<"is between "<<lb<<" and "<<ub<<"\n";
	}
      cout<<endl;

      cout<<"Information on III(E'/Q):\n";

      lb = 1; lb1=1; ub = sha20; eq = (lb==ub);
      cout<<"\t#phi'(III(E/Q)[2]) ";
      if(!eq) cout<<"<";
      cout<<"= " << ub << "\n";

      lb = sel0/sel20; lb1=(lb==1); ub = sha0; eq=(lb==ub);
      cout<<"\t#III(E'/Q)[phi]    ";
      if(eq) cout<<"= "<<ub<<"\n";
      else
	{
	  if(lb1) cout<<"<= "<<ub<<"\n";
	  else    cout<<"is between "<<lb<<" and "<<ub<<"\n";
	}

      lb = sel0/sel20; lb1=(lb==1); ub = sha0*sha21; eq=(lb==ub);
      cout<<"\t#III(E'/Q)[2]      ";
      if(eq) cout<<"= "<<ub<<"\n";
      else
	{
	  if(lb1) cout<<"<= "<<ub<<"\n";
	  else    cout<<"is between "<<lb<<" and "<<ub<<"\n";
	}
      cout<<endl;
#else
      cout<<"Information on III(E/Q):\n";

      ub = sha1;
      if(ub==1) cout<<"\t III(E/Q)[phi'] is trivial\n";
      else cout<<"\t#III(E/Q)[phi'] <= " << ub << "\n";

      ub = sha0*sha1;
      if(ub==1) cout<<"\t III(E/Q)[2]    is trivial\n";
      else cout<<"\t#III(E/Q)[2]    <= " << ub << "\n";

      cout<<"Information on III(E'/Q):\n";

      ub = sha0;
      if(ub==1) cout<<"\t III(E'/Q)[phi] is trivial\n";
      else cout<<"\t#III(E'/Q)[phi] <= " << ub << "\n";

      ub = sha0*sha1;
      if(ub==1) cout<<"\t III(E'/Q)[2]   is trivial\n";
      else cout<<"\t#III(E'/Q)[2]   <= " << ub << "\n";

      cout<<endl;
#endif      
      if(!certain) 
	cout << "(The above may be refinable using parity considerations)\n\n";
    }
  
  if(npoints>0) makegens();
  if(rank<=MAX_R)makepoints();
}

void rank2::makegens()
{
  Curvedata ee_min;
  bigint u, r, s, t, x, y, z; int i, j;
  ee_min=ee.minimalize(u,r,s,t);
  if(verbose)
    {
      cout<<"-------------------------------------------------------\n";
      cout << "\nList of points on E = " << (Curve)ee_min << ":\n";
      cout<<"\nI.  Points on E mod phi(E')\n";
      if(npoints1==0) cout << "--none (modulo 2-torsion).\n\n";
    }
  for(i=0; i<npoints; i++)
    {
      if(verbose&&(i==npoints1)) {cout<<"\nII. Points on phi(E') mod 2E\n";}
      Point q = shift(pointlist[i],the_curve,u,r,s,t);
      bigfloat h = height(q);
      int valid = q.isvalid();
      if(verbose||!valid) cout << "Point " << q << ", height = " << h;
      if(!valid) cout << " --warning: NOT on curve!";
      if(verbose||!valid) cout << "\n";
      pointlist[i]=q;
    }
  if(verbose&&(npoints1==npoints)) 
    {
      cout<<"\nII.  Points on phi(E') mod 2E\n";
      cout << "--none (modulo 2-torsion).\n\n";
    }
}

void rank2::listgens()
{
  long i;
  cout << "List of generators of E(Q)/2E(Q) for E = " 
       << (Curve)(*the_curve) << ": \n";
  for(i=0; i<npoints; i++)
    {
      Point p = pointlist[i];
      cout << "Point " <<
	//	    "on " <<  (Curve)(*CD_orig) << ": " << 
	p;
      bigfloat h = height(p);
      cout << ", height = " << h;
      if(!p.isvalid()) cout << " --warning: NOT on curve!";
      cout << "\n";
    }
}

void rank2::listgens(Curvedata* CD_orig, const bigint& u, const bigint& r, 
		     const bigint& s, const bigint& t)
{
  long i;
  cout << "List of generators of E(Q)/2E(Q) (mod torsion) for E = " 
       << (Curve)(*CD_orig) << ": \n";
  for(i=0; i<npoints; i++)
    {
      Point p = shift(pointlist[i],CD_orig,u,r,s,t,1);
      cout << "Point " << (i+1) <<
	//	    "on " <<  (Curve)(*CD_orig) << 
	": " << p;
      bigfloat h = height(p);
      cout << ", height = " << h;
      if(!p.isvalid()) cout << " --warning: NOT on curve!";
      cout << "\n";
    }
}

void rank2::makepoints()
{
  if(fullnpoints>0) return;   // avoids calling this twice
  int i, j;
  long smallindex = index2/(1+ntwo_torsion);
  fullnpoints=1;         // will be smallindex
  fullpointlist = PointArray(smallindex);
  fullpointlist[0]=Point(the_curve);

  if(verbose&&(rank>0))
    {
      cout<<"-------------------------------------------------------\n";
      cout << "Computing full set of "<<smallindex<<" coset representatives for\n";
      cout << "2E(Q) in E(Q) (modulo torsion), and sorting into height order...." << flush;
    }
  for(i=0; i<rank; i++)
    {
      for(j=0; j<fullnpoints; j++)
	fullpointlist[j+fullnpoints] = fullpointlist[j]+pointlist[i];
      fullnpoints*=2;
    }
  if(fullnpoints!=smallindex)
    {
      cout << "Problem: index = " << index2 << " but " <<fullnpoints<<" cosets\n";
    }
//
// Now reorder points into increasing height order:
//
  for(i=0; i<fullnpoints; i++)
    for(j=i+1; j<fullnpoints; j++)
      if(height(fullpointlist[j])<height(fullpointlist[i]))
	{
	  Point temp = fullpointlist[i]; 
	  fullpointlist[i]=fullpointlist[j]; 
	  fullpointlist[j]=temp;
	}
  if(verbose&&(rank>0))  cout << "done.\n" << endl;
}

void rank2::listpoints()
{
  cout << "Points on curve E = " << (Curve)(*the_curve) 
       << " covering E(Q)/2E(Q), modulo 2-torsion:";
  if(rank==0) cout<<" none.";
  else
    if(rank>MAX_R)
      cout << "Too many to list ("<<fullnpoints<<" mod 2-torsion)\n";
    else
    {
      cout<<"\n"<<fullnpoints<<" points, [0:1:0] and:\n";
      for (long i=1; i<fullnpoints; i++)
	{
	  Point p = fullpointlist[i];
	  cout << "Point " << p;
	  bigfloat h = height(p);
	  cout << ", height = " << h;
	  if(!p.isvalid()) {cout << " --warning: NOT on curve! " ;}
	  cout << "\n";
	}
    }
  cout<<"\n\n";
}

void rank2::listpoints(Curvedata* CD_orig, const bigint& u, const bigint& r, 
		                           const bigint& s, const bigint& t)
{
  cout << "Points on original curve E = " << (Curve)(*CD_orig) 
       << " covering E(Q)/2E(Q), modulo 2-torsion:";
  if(rank==0) cout<<" none.";
  else if(rank>MAX_R) 
    cout << "Too many to list ("<<fullnpoints<<" mod 2-torsion)\n";
  else
    {
      cout<<"\n"<<fullnpoints<<" points:"<<endl;
      for (long i=1; i<fullnpoints; i++)
	{
	  Point p0 = fullpointlist[i];
//	  cout << "Point on " <<  (Curve)(*the_curve) << ": " << p0;
//	  if(!p0.isvalid()) cout << " --warning: NOT on curve!";
//	  cout << "\n";
	  Point p = shift(p0,CD_orig,u,r,s,t,1);
	  cout << "Point " <<
	    //	    "on " <<  (Curve)(*CD_orig) << ": " << 
	    p;
	  bigfloat h = height(p);
	  cout << ", height = " << h;
	  if(!p.isvalid()) cout << " --warning: NOT on curve!";
	  cout << "\n";
	}
    }
  cout<<"\n\n";
}
